//**************************************************************************//
//* FRIDA: fast reliable interactive data analysis                         *//
//* curve.cpp: curve fitting                                               *//
//* (C) Joachim Wuttke 1990-, v2(C++) 2001-                                *//
//* http://frida.sourceforge.net                                           *//
//**************************************************************************//

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <string>
#include <vector>
#include <algorithm>

#include "olm.h" // includes also curve.h
#include "asi.h"
#include "../../lmfit/lmmin.h"
#include "xax.h"

using namespace std;


//***************************************************************************//
//* global namespace: lm_minimize callback functions                        *//
//***************************************************************************//


//! Data transfer between frida and lm_minimize callback functions.

typedef struct {
    COld *fd;
    COlc *fc;
    uint k;
    uint j;
} GlobalDataType;


//! Callback routine for use in lm_minimize: function evaluation.

void globalEvaluate( double* par, int m_dat, double* fvec, 
                     void *data, int *info )
// arguments (required by lm_minimize):
//    par   r/w  function parameters, to be optimized
//    m_dat r    number of data points
//    fvec  w    vector of residues, to be minimized
//    data  r    pointer to arbitrary data, here defined as GlobalDataType
//    info  w    can be use to terminate the minimization
{
    GlobalDataType *mydata;
    mydata = (GlobalDataType*) data;
    COlc *fc = mydata->fc;
    COld *fd = mydata->fd;
    CCurve *c = &(fc->VC[ mydata-> j ]);

    uint ip = 0;
    for( uint i=0; i<fc->nPar(); ++i )
        if ( !(fc->Fixed[i]) )
            c->P[i] = par[ip++];

    double fit;
    CScan *s = &(fd->VS[mydata->j]);
    for ( uint i=0; i<m_dat; i++) {
        fit = fc->T.tree_curve_val( s->x[i], fc, mydata->k, mydata->j, i);
        if (!fc->weight_y_log) {
            fvec[i] = s->y[i] - fit;
        } else {
            if ( s->y[i] < 1e-15 || fit < 1e-15 ) {
                fvec[i] = 0; // just don't fit
            } else {
                fvec[i] = log( s->y[i] / fit );
            }
        }
    }
}


//! Callback routine for use in lm_minimize: monitor fit progress.

void globalPrintout( int n_par, double* par, int m_dat, double* fvec, 
                     void *data, int iflag, int iter, int nfev )
// arguments (required by lm_minimize; see also globalEvaluate):
//       iflag : 0 (init) 1 (outer loop) 2(inner loop) -1(terminated)
//       iter  : outer loop counter
//       nfev  : number of calls to *evaluate
{
}


//***************************************************************************//
//* class CFitting                                                          *//
//* (this class is basically a namespace with a few parameters)             *//
//***************************************************************************//


//! On program upstart, set default parameters.

CFitting::CFitting( void )
    : nCall(200), nPrint(0), Factor(100.), Epsilon(1.e-14), mFitMetric(1)
{
}


//! Set auxiliary parameters. Directly called from main program.

void CFitting::SetSetup( string which )
{
    if      ( which=="p" )
        nPrint = iask( "Print frequency", nPrint );
    else if ( which=="c" )
        nCall = iask( "Max no of calls", nCall );
    else if ( which=="f" )
        Factor = dask( "Factor", Factor );
    else if ( which=="e" )
        Epsilon = dask( "Epsilon", Epsilon );
    else if ( which=="m" ){
        // metric (0) is the integer indicator returned by lm_minimize;
        // it cannot be chosen here
        printf( "Fit metric:\n"
                "(1) coefficient of determination\n"
                "(2) mean deviation\n" );
        mFitMetric = iask_lim( "Option", mFitMetric,
                               CLimits( 1, N_FIT_METRICS-1 ) );
    }
}


//! Create curve files.

void CFitting::Create(void)
{
    COld *fin;
    CTree T;

    string expr = sask("Fit function ?");
    if (expr==string("")) return;
    if ( user_xaxparse(expr.c_str(), &T) ) {
        printf("! invalid expression\n");
        return;
    }
    // cout << "info: " << T.info() << endl;

    uint np = T.npar(CRef::_FP);

    // create one curve file for each selected data file
    NOlm::IterateD fiter;
    NOlm::SelNew();
    while((fin=fiter())) {
        COlc fout(fin);
        fout.expr = expr;
        fout.T = T;
        fout.name = string("fit_") + fin->name;
        fout.kd = fiter.SelNo();
        fout.weight_y_log = false;
        for (uint ip=0; ip<np; ip++) {
            fout.PCo.push_back(CCoord("p" + strg(ip), ""));
            fout.Fixed.push_back( 0 );
        }
        for (uint j=0; j<fin->nScan(); j++) {
            CCurve sout;
            sout.Clear();
            sout.z = fin->VS[j].z;
            for (uint k=0; k<fout.nPar(); k++) {
                sout.P.push_back(1.);
            }
            fout.VC.push_back(sout);
        }
        NOlm::OloAdd(&fout);
    }
}


//! Set file-dependent auxiliary parameters. Directly called from main.

void CFitting::SetProperties(string which)
{
    COlc *fc;
    NOlm::IterateC fiter;
    while((fc=fiter())) {
        if (which=="yli")
            fc->weight_y_log = false;
        else if (which=="ylg")
            fc->weight_y_log = true;
        else {
            printf("! unknown property");
            return;
        }
    }
}


//! Fit.

void CFitting::Fit(void)
{
    COlc *fc;
    COld *fd;
    CScan *D;
    CCurve *C;
    vector<double> Par;
    lm_control_type control;
    GlobalDataType data;
    double ysum1, ysum2, ydev1, ydev2;

    NOlm::IterateC fiter;
    while((fc=fiter())) {
        data.k = fiter.SelNo();
        fc->kd = iask("Fit data file no.", fc->kd);
        if (!(fd=NOlm::MOM[fc->kd].D())) {
            printf("! is not a data file\n");
            return;
        }
        if (fc->nScan() != fd->nScan()) {
            printf("! data file and curve file have "
                   "different number of scans\n");
            return;
        }

        data.fc = fc;
        data.fd = fd;

        uint nx = fc->T.npar(CRef::_X);
        if (nx<=0) {
            printf("! function must depend on x\n");
            return;
        }

        uint np = fc->T.npar(CRef::_FP);

        // determine number of free parameters
        uint npfree = 0;
        for ( uint ip=0; ip<np; ++ip )
            if( !fc->Fixed[ip] )
                ++npfree;
        if (npfree<=0) {
            printf("! function must depend on some free "
                   "p-parameters\n");
            return;
        }
        Par.resize(npfree);
                
        lm_initialize_control( &control );
        control.epsilon = Epsilon;
        control.stepbound = Factor;
        control.maxcall = nCall;

        for (uint j=0; j<fc->nScan(); j++) {
            data.j = j;
            printf("%3d", j);
            D = &(fd->VS[j]);
            C = &(fc->VC[j]);
            uint ipf=0;
            for ( uint ip=0; ip<np; ip++ ) {
                if ( !fc->Fixed[ip] )
                    Par[ipf++] = C->P[ip];
            }
                                        
            printf( " [%d*%d]", npfree, D->size() );
            if ( nPrint ) printf("\n");

            lm_minimize( D->size(), npfree, &(Par[0]), 
                         globalEvaluate, globalPrintout,
                         &data, &control );

            printf( " %4d", control.nfev);
            for ( uint i=0; i<npfree; i++ ) printf(" %12g", Par[i]);

            C->Quality[0] = (double) control.info;

            // calculate coefficient of determination R^2:
            uint nd = fd->nPts(j);
            ysum1 = 0, ysum2 = 0;
            for( uint i=0; i<nd; ++i ){
                ysum1 += D->y[i];
                ysum2 += SQR( D->y[i] );
            }
            ydev2 = ysum2 - SQR( ysum1 )/nd; // = nd * <(y-<y>)^2>
            C->Quality[1] = ydev2!=0 ? 1 - SQR(control.fnorm) / ydev2 : -1;
                        
            // mean deviation:
            C->Quality[2] = SQR(control.fnorm) / nd;

            printf( " (%9.7g)", C->Quality[mFitMetric] );

            // print status:
            printf( "%s\n", lm_shortmsg[ control.info ] );
        }
    }
}


//! Store fit parameters as integral file.

void CFitting::GetPar(void)
{
    NOlm::IterateC fiter;
    NOlm::SelNew();
    COlc *fin;
    static CList parsel;

    int maxpar = 0;
    while((fin=fiter())) {
        if( fin->nPar() > maxpar ) maxpar = fin->nPar();
    }
    if( maxpar<1) return;
    parsel.Ask( "Save which parameters", parsel, CLimits( 0, maxpar ) );

    fiter.Reset();
    while((fin=fiter())) {
        uint nz = fin->ZCo.size();
        bool savable = ( nz > 0 );
        printf(" savable: %c\n", savable ? 'y' : 'n' );

        for ( int iv=0; iv<parsel.V.size(); ++iv ) {
            uint ip = parsel.V[iv];
            if( ip>=fin->nPar() ) continue;

            COld fout(fin);

            CCoord zco = fin->ZCo.back();
            fout.ZCo.pop_back();
            fout.xco = zco;
            fout.yco = fin->PCo[ip];
		
            CScan S;
            S.z = fin->VC[0].z;
            S.z.pop_back();
            for (uint j=0; j<fin->nScan(); j++) {
                if (!savable) {
                    printf( "par%2d of scan%d has value %12g\n", 
                            ip, j, fin->VC[0].P[ip]); 
                    continue;
                }
                
                S.push_back( fin->VC[j].z[nz-1], fin->VC[j].P[ip] );
                if (nz>=2 &&
                    (j+1)<fin->nScan() && 
                    fin->VC[j+1].z[nz-2]!=fin->VC[j].z[nz-2]) {
                    fout.VS.push_back(S);
                    S.Clear();
                    S.z = fin->VC[j+1].z;
                    S.z.pop_back();
                }
            }
            if (savable) {
                fout.VS.push_back(S);
                NOlm::OloAdd(&fout);
            }
        }
    }
};


//! Display list of fit parameters.

void CFitting::ShowPar(void)
{
    printf( "Fit setup: nCall=%d, nPrint=%d, Factor=%g, Eps=%g\n",
            nCall, nPrint, Factor, Epsilon );
    NOlm::IterateC fiter;
    COlc *fin;
    while((fin=fiter())) {
        printf( "curve %d to fit data %d: %s\n", 
                fiter.SelNo(), fin->kd, fin->expr.c_str() );
        printf( "weight(ylog)=%d\n", fin->weight_y_log );
        for (uint ip=0; ip<fin->nPar(); ip++) {
            printf( "p%2d %s        ", ip,
                    fin->Fixed[ip] ? "fix" : "   " );
        }
        printf("\n");
        for (uint j=0; j<fin->nScan(); j++) {
            for (uint ip=0; ip<fin->nPar(); ip++)
                printf( "%12.7g ", fin->VC[j].P[ip]);
            printf("\n");
        }
    }
}


//! Overwrite fit parameters.

void CFitting::OprPar(string llabel) 
{
    CRef lref(llabel);
    if (!lref.defined()) {
        printf("! invalid left-hand side %s\n", llabel.c_str());
        return;
    } else if (lref.var!=CRef::_FP) {
        printf("PROGRAM ERROR: OprPar: im falschen Film\n");
        return;
    }

    string expr = sask("Function ?");
    if (expr=="") return;
    CTree T;
    if ( user_xaxparse(expr.c_str(), &T) ) {
        printf("! invalid expression\n");
        return;
    }

    NOlm::IterateC fiter;
    COlc *fin;
    double dout;
    while((fin=fiter())) {
        if( lref.num>=fin->nPar() ){
            printf("! invalid left-hand side: not that many parameters\n" );
            return;
        }
        for (uint j=0; j<fin->nScan(); j++) {
            // printf("DEBUG .. operate on scan %d\n", j);

            if (T.tree_val(&dout, (COlo*) fin,
                           fiter.SelNo(), j)) 
                return;
            fin->VC[j].P[lref.num] = dout;
        }
    }
}


//! Fix or unfix fit parameters.

void CFitting::SetFixed(string how) 
{

    const char *how_opts[] = { "", "+", "-" };
    int ihow;
    for ( ihow=0; ihow<=2; ++ihow )
        if ( how==how_opts[ihow] )
            goto found_how;
    cout << "ERROR/ unknown option\n";
    return;
 found_how:

    const char *which_quest[] = { 
        "toggle fix status of pars", "set fixed pars", "unfix pars" };

    NOlm::IterateC fiter;
    COlc *fin;
    while((fin=fiter())) {

        CList Which;
        Which.Ask( which_quest[ihow], CLimits(0,fin->nPar()-1) );

        IListV<int> Piter(&(fin->Fixed), &Which);
        int *p;
        switch (ihow) {
        case 0: while ((p=Piter())) *p = !*p;   break;
        case 1: while ((p=Piter())) *p = 1;  break;
        case 2: while ((p=Piter())) *p = 0; break;
        }
    }
}


//***************************************************************************//
//* class CCurve                                                            *//
//* (one curve file may contain several curves, just as one data file may   *//
//*  contain several scans)                                                 *//
//***************************************************************************//


//! Initialize curve. Used when creating or loading a curve file.

void CCurve::Clear(void)
{
    P.clear();
    z.clear();
    Quality[0] = -1;
    Quality[1] = -1;
    Quality[2] = -1;
}


//! Callback, for saving curve files.

int CCurve::Save(FILE *fd)
{
    printf("WARNING: curve saving not yet implemented\n");
    return 1;
//	return 0;
}


//! Callback, for loading curve files.

int CCurve::Load(FILE *fd, uint nz)
{
    printf("WARNING: curve loading not yet implemented\n");
    Clear();
    return 0;
}
