//**************************************************************************//
//* FRIDA: fast reliable interactive data analysis                         *//
//* expr.cpp: evaluate generic expressions                                 *//
//* (C) Joachim Wuttke 1990-, v2(C++) 2001-                                *//
//* http://frida.sourceforge.net                                           *//
//**************************************************************************//

#include <stdlib.h>
#include <ctype.h>
#include <math.h>

#include "olm.h" // for cross-references

typedef struct {
    COlc* fc;
    uint j;
} CurveRef;


//***************************************************************************//
//* xax: a flex/bison based parser                                          *//
//***************************************************************************//

#include "xax_yacc.h"
#include "xax.h"

int xaxerror( const char *s )  /* Called by yyparse on error */
{
    printf ("xaxerror: %s\n", s);
    return 1;
}


//***************************************************************************//
//* class CRef                                                              *//
//* (references, typically used as endpoints of CTree's)                    *//
//***************************************************************************//

CRef::CRef(string s, 
	   class CTree *tk_in, class CTree *tj_in, class CTree *ti_in)
    // NOTE: As long as this constructor is called by the parser and not
    // by the lexical analyzer, string s may contain more than just the
    // parameter name. Everything after the parameter name must be ignored.
{
    // printf( "DEBUG CRef s[%s]\n", s.c_str() );
    num = 0;
    int isub;
    char c;
    int nsub = sscanf(s.c_str(), "%c%u", &c, &isub);
    if (s.substr(0,2)=="fm" && sscanf(s.substr(2).c_str(), "%u", &isub)==1 ) {
        var = _FM; // fit metric
        num = isub;
    } else if (nsub==2 && c=='p') {
        var = _FP; // fit parameter
        num = isub;
    } else if (nsub==2 && c=='z') {
        var = _Z;  // z value
        num = isub;
    } else if (s.substr(0,2)=="nj") {
        var = _NJ; // number of scans in file
    } else if (s.substr(0,2)=="ni") {
        var = _NI; // number of points in scan
    } else if (s.substr(0,2)=="ak") {
        var = _AK; // ask file-by-file
    } else if (s.substr(0,2)=="aj") {
        var = _AJ; // ask scan-by-scan
    } else if (s.substr(0,2)=="ai") {
        var = _AI; // ask point-by-point
    } else if (nsub==1 && c=='x') {
        var = _X;  // x value
    } else if (nsub==1 && c=='y') {
        var = _Y;  // y value
    } else if (nsub==1 && c=='f') {
        var = _FC; // fit curve
    } else if (nsub==1 && c=='k') {
        var = _K;  // internal file index
    } else if (nsub==1 && c=='j') {
        var = _J;  // index of current scan
    } else if (nsub==1 && c=='i') {
        var = _I;  // index of current point
    } else {
        // printf( "DEBUG found noref [%s]\n", s.c_str() );
        var = _NOREF;
    }
    if (isask()) {
        if (sscanf(s.substr(2).c_str(), "%u", &isub)==1)
            num = isub;
        else
            num = 0;
    }
    // newly constructed CRef takes over responsability to destroy CTree's
    tk = tk_in;
    tj = tj_in;
    ti = ti_in;
    if (isindex() && indexed()) {
        printf("! invalid cross reference: index cannot be indexed\n");
        var = _NOREF;
        return;
    }
}

CRef::~CRef()
{
    if (tk) delete tk;
    if (tj) delete tj;
    if (ti) delete ti;
}

CRef& CRef::operator=(const CRef& R2)
{
    var = R2.var;
    num = R2.num;
    if (R2.tk) { tk = new CTree (); *tk = *(R2.tk); } else tk=0;
    if (R2.tj) { tj = new CTree (); *tj = *(R2.tj); } else tj=0;
    if (R2.ti) { ti = new CTree (); *ti = *(R2.ti); } else ti=0;
    return *this;
}

int CRef::ref_val(void *ret, CContext *ctx ) const
{
    uint i,j,k;
    COlo *f;
    COld *fd;
    COlc *fc;
    CurveRef* cr;

    // printf( "DEBUG ref_val\n" );
    if (isindex()) {
        switch(var) {
        case CRef::_K: 
            *((double*) ret) = ctx->k;
            return 0;
        case CRef::_J: 
            *((double*) ret) = ctx->j;
            return 0;
        case CRef::_I: 
            *((double*) ret) = ctx->i;
            return 0;
        default:
            printf("PROGRAM ERROROR ref_val(index)\n");
            return -2;
        }
    }

    if (ti) {
        // get i from reference :
        if (!pointwise()) {
            printf( "! index i incompatible with ref %s\n", ref_cstr() );
            return -1;
        }
        if (ti->tree_uival( &i, ctx ) ) {
            printf( "! could not eval index i=%s\n", ti->tree_cstr() );
            return -1;
        }
    } else {
        i = ctx->i;
    }

    if (tj) {
        // get j from reference :
        if (!scanwise()) {
            printf( "! index j incompatible with ref %s\n", ref_cstr() );
            return -1;
        }
        if (tj->tree_uival( &j, ctx ) ) {
            printf( "! could not eval index j=%s\n", tj->tree_cstr() );
            return -1;
        }
    } else {
        j = ctx->j;
    }
		
    if (tk) {
        // get k from reference :
        if (tk->tree_uival( &k, ctx ) ) {
            printf( "! could not eval index k=%s\n", tk->tree_cstr() );
            return -1;
        }
        if (k>=NOlm::MOM.size()) {
            printf("! file index %u invalid, maximum is %u\n",
                   k, NOlm::MOM.size()-1);
            return -1;
        }
        f = NOlm::MOM[k].P();
    } else {
        f = ctx->f;
    }
		
    if (!f) {
        printf("! ref where not referring to a file\n");
        return -1;
    }
    if (pointwise() && i==(uint)-1) {
        printf("! ref(point) out of context\n");
        return -1;
    }
    if (scanwise() && j==(uint)-1) {
        printf("! ref(scan) out of context\n");
        return -1;
    }

    if ((fd = dynamic_cast<COld*>(f))) {
        if (!indatafile()) {
            printf("! ref not allowed in data file\n");
            return -1;
        }
        if (scanwise() && j>=fd->nScan()) {
            printf("! ref(scan) out of range\n");
            return -1;
        }
        if (pointwise() && i>=fd->VS[j].size()) {
            printf("! ref(point) out of range\n");
            return -1;
        }
        switch(var) {
        case CRef::_X: 
            *((double*) ret) = fd->VS[j].x[i];
            return 0;
        case CRef::_Y:
            *((double*) ret) = fd->VS[j].y[i];
            return 0;
        case CRef::_Z:
            if( num>= fd->nZ() ){
                printf( "!!! invalid z ref(%s) in data file\n", ref_cstr() );
                return -2;
            }
            *((double*) ret) = fd->VS[j].z[num];
            return 0;
        case CRef::_NJ: 
            *((double*) ret) = fd->nScan();
            return 0;
        case CRef::_NI: 
            *((double*) ret) = fd->VS[j].size();
            return 0;
        default:
            printf("!!! invalid ref(%s) in data file\n",
                   ref_cstr());
            return -2;
        }
    } else if ((fc = dynamic_cast<COlc*>(f))) {
        if (!incurvefile()) {
            printf("! ref not allowed in curve file\n");
            return -1;
        }
        if (scanwise() && j>=fc->nScan()) {
            printf("! ref(scan) out of range\n");
            return -1;
        }
        switch(var) {
        case CRef::_FP:
            if( num>= fc->nPar() ){
                printf("!!! invalid p ref(%s) in curve file\n",
                       ref_cstr());
                return -2;
            }
            *((double*) ret) = fc->VC[j].P[num];
            // printf( "DEBUG p%i[%i]=%g\n", num, j, *((double*) ret) ); 
            return 0;
        case CRef::_FM:
            if( num>= N_FIT_METRICS ){
                printf( "!!! invalid fm ref(%s) in curve file\n", ref_cstr() );
                return -2;
            }
            *((double*) ret) = fc->VC[j].Quality[num];
            return 0;
        case CRef::_Z:
            if( num>= fc->nZ() ){
                printf( "!!! invalid z ref(%s) in curve file\n", ref_cstr() );
                return -2;
            }
            *((double*) ret) = fc->VC[j].z[num];
            return 0;
        case CRef::_FC:
            cr = (CurveRef*) ret;
            cr->fc = fc;
            cr->j = j;
            return 0;
        case CRef::_NJ: 
            *((double*) ret) = fc->nScan();
            return 0;
        default:
            printf( "!!! invalid Ref(%s) in curve file\n", ref_cstr() );
            return -2;
        }
    } else {
        printf("PROGRAM ERROR ref->eval unexpected else\n");
        return -3;
    }
}

int CRef::coord( CCoord *ret, class COlo* f_in, uint k_in ) const
{
    COlo *f;
    COld *fd;
    COlc *fc;
    uint k;
    CContext ctx( f_in, k_in, 0, 0 );

    if (tk) {
        // get k from reference :
        if (tk->tree_uival(&k, &ctx)) {
            printf( "! could not eval index k=%s\n", tk->tree_cstr() );
            return -1;
        }
        if (k>=NOlm::MOM.size()) {
            printf( "! file index %u invalid, maximum is %u\n",
                    k, NOlm::MOM.size()-1);
            return -1;
        }
        f = NOlm::MOM[k].P();
    } else {
        f = f_in;
        k = k_in;
    }

    if (!f) {
        printf("! ref where not refering to a file\n");
        return -1;
    }
    if (isindex()) {
        switch(var) {
        case CRef::_K: 
            *ret = CCoord("k","");
            return 0;
        case CRef::_J: 
            *ret = CCoord("j","");
            return 0;
        case CRef::_I: 
            *ret = CCoord("i","");
            return 0;
        default:
            printf("PROGRAM ERROROR coord(index)\n");
            return -2;
        }
    } else if ((fd = dynamic_cast<COld*>(f))) {
        if (!indatafile()) {
            printf("! coord ref %s not allowed in data file %i\n",
                   ref_cstr(), k);
            return -1;
        }
        switch(var) {
        case CRef::_X: 
            *ret = fd->xco;
            return 0;
        case CRef::_Y:
            *ret = fd->yco;
            return 0;
        case CRef::_Z:
            if (num>=fd->ZCo.size()) {
                printf("! z%u undefined\n", num);
                return -1;
            }
            *ret = fd->ZCo[num];
            return 0;
        case CRef::_NJ:
            *ret = CCoord("#scans", "");
            return 0;
        case CRef::_NI:
            *ret = CCoord("#entries", "");
            return 0;
        default:
            printf("!!! coord data:ref(%s) nyi\n",
                   ref_cstr());
            return -2;
        }
    } else if ((fc = dynamic_cast<COlc*>(f))) {
        if (!incurvefile()) {
            printf( "! coord ref %s not allowed in curve file\n", ref_cstr() );
            return -1;
        }
        switch(var) {
        case CRef::_FP:
            *ret = fc->PCo[num];
            return 0;
        case CRef::_FM:
            *ret = CCoord("fit_quality_indicator_"+strg(num), "");
            return 0;
        case CRef::_Z:
            *ret = fc->ZCo[num];
            return 0;
        case CRef::_FC:
            *ret = CCoord("function","");
            return 0;
        case CRef::_NJ:
            *ret = CCoord("#scans", "");
            return 0;
        default:
            printf( "!!! coord curv:ref(%s) nyi\n", ref_cstr() );
            return -2;
        }
    } else {
        printf("PROGRAM ERROR ref->coord unexpected else\n");
        return -3;
    }
}

string CRef::ref_info(void) const
{
    string s;
    switch (var) {
    case _X:
        s = "x";
    case _Y:
        s = "y";
    case _FC:
        s = "f";
    case _Z:
        s = "z";
    case _FP:
        s = "p";
    case _FM:
        s = "fm";
    case _K:
        s = "k";
    case _J:
        s = "j";
    case _I:
        s = "i";
    case _NJ:
        s = "nj";
    case _NI:
        s = "ni";
    default:
        s = "?";
    }
    if (numbered()) s += strg(num);
    return s;
}


//***************************************************************************//
//* class CTree                                                             *//
//***************************************************************************//

CTree& CTree::operator=(const CTree& T2)
{
    typ = T2.typ;
    switch (typ) {
    case _TEROP: // ternary function
        arg3 = new CTree();
        *arg3 = *(T2.arg3);
    case _BINOP: // binary function
        arg2 = new CTree();
        *arg2 = *(T2.arg2);
    case _MONOP: // unary function
        arg1 = new CTree();
        *arg1 = *(T2.arg1);
        fun = T2.fun;
        break;
    case _OP_T_V: // binary function of tree and vector<tree>: "i in [1,2,4]"
        arg1 = new CTree();
        *arg1 = *(T2.arg1);
        varg2 = new VCTree();
        *varg2 = *(T2.varg2);
        fun = T2.fun;
        break;
    case _VAL: // numeric value
        val = T2.val;
        break;
    case _FRF: // function reference
        arg1 = new CTree();
        *arg1 = *(T2.arg1);
    case _REF: // reference
        ref = new CRef();
        *ref = *(T2.ref);
        break;
    default:
        printf("PROGRAM ERROR: CTree operator=: assign undef'd tree\n");
        scanf("\n");
    }
    return *this;
}


//! Evaluate tree and convert result to integer, for use as index.

int CTree::tree_uival( uint *ret, CContext *ctx ) const
{
    double aux;
    if (tree_val( &aux, ctx )) return -1;
    // since we have no integer data type, convert real by rounding
    *ret = (uint) (aux + 0.5);
    return 0;
}


//! Evaluate tree.

int CTree::tree_val( double *ret, CContext *ctx ) const
{
    double r1, r2, r3;
    vector<double> v2;
    CContext *cu_ctx;

    switch (typ) {
    case _MONOP:
        if( arg1->tree_val( &r1, ctx ) ){
            printf( "! tree could not eval 1st arg of %s\n", tree_cstr() );
            return -1;
        }
        *ret = (*(fun->f1))(r1);
        return 0;
    case _BINOP:
        if( arg1->tree_val( &r1, ctx ) ){
            printf( "! tree could not eval 1st arg of %s\n", tree_cstr() );
            return -1;
        }
        if( arg2->tree_val( &r2, ctx ) ){
            printf( "! tree could not eval 2nd arg of %s\n", tree_cstr() );
            return -1;
        }
        *ret = (*(fun->f2))(r1, r2);
        return 0;
    case _TEROP:
        if( arg1->tree_val( &r1, ctx ) ){
            printf( "! tree could not eval 1st arg of %s\n", tree_cstr() );
            return -1;
        }
        if( arg2->tree_val( &r2, ctx ) ){
            printf( "! tree could not eval 2nd arg of %s\n", tree_cstr() );
            return -1;
        }
        if( arg3->tree_val( &r3, ctx ) ){
            printf( "! tree could not eval 3rd arg of %s\n", tree_cstr() );
            return -1;
        }
        *ret = (*(fun->f3))(r1, r2, r3);
        return 0;
    case _OP_T_V:
        if( arg1->tree_val( &r1, ctx ) ){
            printf( "! tree could not eval 1st arg of %s\n", tree_cstr() );
            return -1;
        }
        for( uint iv2=0; iv2<varg2->size(); ++iv2 ){
            if ((*varg2)[iv2]->tree_val( &r2, ctx ) ){
                printf( "! tree could not eval %dth elem of 2nd arg of %s\n",
                       iv2, tree_cstr() );
                return -1;
            }
            v2.push_back( r2 );
        }
        *ret = (*(fun->f2tv)) ( r1, v2 );
        return 0;
    case _VAL:
        *ret = val;
        return 0;
    case _REF:
        if( ctx->flags & CContext::HAS_XARG ){
            switch(ref->var) {
            case CRef::_X:
                *ret = ctx->xarg;
                return 0;
            case CRef::_FP:
                if( ctx->j==-1 ){
                    printf( "PROGRAM ERROR: "
                            "missing j in tree_curve_val:REF:FP\n" );
                    return -1;
                }
                *ret = ((COlc*)ctx->f)->VC[ctx->j].P[ref->num];
                return 0;
            case CRef::_NOREF: 
                printf("PROGRAM ERROR:"
                       " tree_curve_val(ref_typ=NOREF)\n");
                return -1;
            default:
                printf("PROGRAM ERROR:"
                       " tree_curve_val(ref_typ=?)\n");
                return -1;
            }
        } else {
            if (ref->ref_val( (void*)ret, ctx ) ){
                printf( "! tree could not eval %s\n", tree_cstr() );
                return -1;
            }
            return 0;
        }
    case _FRF:
        if( arg1->tree_val( &r1, ctx ) ){
            printf( "! tree could not eval external function arg of %s\n",
                    tree_cstr() );
            return -1;
        }
        CurveRef cr;
        if ( ref->ref_val((void*)&cr, ctx ) ){ 
            printf( "! tree could not eval external function ref"
                    " of %s [%d,%d,%d](%f)\n", tree_cstr(), 
                    ctx->k, ctx->j, ctx->i, r1 );
            return -1;
        }
        cu_ctx = new CContext(
            cr.fc, ctx->k, cr.j, ctx->i, CContext::HAS_XARG, r1 );
        cr.fc->T.tree_val( ret, cu_ctx ); 
        delete cu_ctx;
        return 0;
    case _NOTREE:
        printf("PROGRAM ERROR: tree_val(tree_typ=NOTREE)\n");
        return -1;
    default:
        printf("PROGRAM ERROR: tree_val(tree_typ=UNDEF)\n");
        return -1;
    }
}


//! Evaluate tree. Explicit API.

int CTree::tree_val(
    double *ret, COlo *f, uint k, uint j, uint i, uint flags, double xarg
    ) const
{
    CContext ctx( f, k, j, i );
    return tree_val( ret, &ctx );
}


//! Evaluate a curve file's tree, substituting unindexed _X and _FP.

double CTree::tree_curve_val(
    double xarg, COlc* fc, uint k, uint j, uint i ) const
{
    double ret;
    CContext ctx( (COlo*)fc, k, j, i , CContext::HAS_XARG, xarg );
    tree_val( &ret, &ctx );
    return ret;
}


//! Compose coordinate names and units according to the tree.

int CTree::coord( CCoord *ret, COlo *f, uint k ) const
{
    CCoord r1, r2, r3;
    if ( fun && fun->narg>=1 && arg1->coord(&r1,f,k)) {
        printf("! tree misses coord of 1st arg of %s\n", tree_cstr());
        return -1;
    }
    if ( fun && fun->narg>=2 && arg2->coord(&r2,f,k)) {
        printf("! tree misses coord of 1st arg of %s\n", tree_cstr());
        return -1;
    }
    if ( fun && fun->narg>=3 && arg3->coord(&r3,f,k)) {
        printf("! tree misses coord of 1st arg of %s\n", tree_cstr());
        return -1;
    }
    switch (typ) {
    case _MONOP:
        *ret = fun->coord(&r1);
        return 0;
    case _BINOP:
        *ret = fun->coord(&r1, &r2);
        return 0;
    case _TEROP:
        *ret = fun->coord(&r1, &r2, &r3);
        return 0;
    case _OP_T_V:
        *ret = CCoord("IN", "");
        return 0;
    case _VAL:
        *ret = CCoord(strg(val), "");
        return 0;
    case _REF:
        if (ref->coord(ret,f,k)) {
            printf("! tree misses coord of %s\n", tree_cstr());
            return -1;
        }
        return 0;
    case _FRF:
        *ret = CCoord("f", ""); // solange j-Querverweis unmglich ...
        return 0;
        if (ref->coord(ret,f,k)) {
            printf("! tree misses coord of %s\n", tree_cstr());
            return -1;
        }
        return 0;
    case _NOTREE: default:
        printf("PROGRAM ERROR: tree::coord(tree_typ=UNDEF)\n");
        return -1;
    }
}

string CTree::tree_info() const
{
    switch (typ) {
    case _MONOP:
        return "mop[" + arg1->tree_info() + "]";
    case _BINOP:
        return "bop[" + arg1->tree_info() + "," + arg2->tree_info() + "]";
    case _TEROP:
        return "3op[" + arg1->tree_info() + "," + arg2->tree_info() + 
            "," + arg3->tree_info() + "]";
    case _OP_T_V:
        return "optv[" + arg1->tree_info() + ", [..] ]";
    case _VAL:
        return "val[" + strg(val) + "]";
    case _REF:
        return "ref[" + ref->ref_info() + "]";
    case _FRF:
        return "frf[" + ref->ref_info() + "]";
    case _NOTREE: default:
        return "notree (PROGRAM ERROR ?)";
    }
}

uint CTree::npar( CRef::TVar v ) const
{
    uint np=0; // number of instances of variable v# 
    nparx(v, &np);
    return np;
}

void CTree::nparx( CRef::TVar v, uint *np ) const
{
    switch (typ) {
    case _MONOP:
        arg1->nparx(v, np);
        return;
    case _BINOP:
        arg1->nparx(v, np);
        arg2->nparx(v, np);
        return;
    case _TEROP:
        arg1->nparx(v, np);
        arg2->nparx(v, np);
        arg3->nparx(v, np);
        return;
    case _OP_T_V:
        arg1->nparx(v, np);
        for( uint i=0; i<varg2->size(); ++i )
            (*varg2)[i]->nparx(v, np);
        return;
    case _VAL:
        return;
    case _REF:
        if (ref->var==v) {
            if (ref->numbered())
                *np = max (*np, ref->num+1);
            else
                *np = 1;
        }
        return;
    case _NOTREE: default:
        printf("PROGRAM ERROR: npar(UNDEF)\n");
        return;
    }
}


//***************************************************************************//
//* direct access from main program                                         *//
//***************************************************************************//

#include "asi.h"
#include "readln.h"

void NCalc::Calculator() 
{
    string s;
    CTree T;
    double v;

    // if there is an expression on the stack, then
    // evaluate just this single expression;
    // otherwise start a loop and request input
    bool expr_on_stack = NRead::nonempty();

    while (1) {
        if( expr_on_stack ){
            s = NRead::readln("");
        } else {
            s = sask("calc> ");
            if (s[0]=='q') return;
            if (s[0]=='h' || s[0]=='?') {
                printf(
                    "enter valid arithmetic expression, or q to quit\n");
                continue;
            }
        }
        if (user_xaxparse(s.c_str(), &T)) {
            printf("! invalid expression\n");
        } else if (!(T.tree_val(&v))) { 
               // if return value is not equal 0, do nothing,
               // because error message has already been printed
            printf("%.12g\n", v);
        }
        if( expr_on_stack ) break;
    }
}
