//**************************************************************************//
//* FRIDA: fast reliable interactive data analysis                         *//
//* grid.cpp: more compact storage of regularly spaced x-data              *//
//* (C) Joachim Wuttke 1990-, v2(C++) 2001-                                *//
//* http://frida.sourceforge.net                                           *//
//**************************************************************************//

#include <iostream>
#include <math.h>
#include "mystd.h"
#include "range.h"
#include "asi.h"

#include "grid.h"

CGrid::CGrid(void)
{
	kind = _NOGRID;
	v.resize(0);
}

CGrid::CGrid(const vector<double>& v_in)
{
	kind = _VEC;
	v = v_in;
}

CGrid::CGrid(const double* v_in, const size_t n_in)
{
	kind = _VEC;
	v = vector<double>(n_in);
	for (uint i=0; i<n_in; ++i)
		v[i] = v_in[i];
}

CGrid::CGrid(const uint n_in, const double start, const double step)
{
	*this = CGrid(n_in, CRange(start, start+(n_in-1)*step));
}

CGrid::CGrid(const uint n_in, const CRange& R, GridTyp k_in)
{
	switch(k_in) {
	case _LIN: case _LOG:
		kind = k_in;
		break;
	default:
		printf("!!! CGrid constructor with invalid GridTyp\n");
		kind = _NOGRID;
		return;
	}
		
	v.resize(0);

	if (n_in==0) {
		printf("CGRID NOGRID\n");
		kind = _NOGRID;
		return;
	}

	if (kind==_LOG  && (R.inf<=0 || R.sup<=0)) {
		printf("!! Cannot construct log grid with limit <= 0\n");
		kind = _NOGRID;
		return;
	}

	ng = n_in;
	gi = R.inf;
	gf = R.sup;
}

void CGrid::simplify(void)
{
	if (kind==_VEC)	try_lin();
	// if (kind==_VEC)	try_log();
}

void CGrid::try_lin(void)
{
	if (kind!=_VEC) return;
	double di, dd, df;
	uint n = v.size();
	if (n==0) {
		CGrid(0);
		return;
	}
	di = v[0];
	df = v[n-1];
	dd = (df-di) / (n-1);
	double dmax = 0, step;
	uint imax = 0;
	for (uint i=1; i<n; i++) {
		// printf("DEB %4u %12g %12g %12g %12g\n", 
		//        i, v[i], di+i*dd, di, dd);
		// if (!eq(v[i], di+i*dd, dd)) return;
		if ((step=fabs(v[i]-v[i-1]-dd)) > dmax) {
			dmax = step;
			imax = i;
		}
	}
	if (dmax>grid_tol*fabs(df-di)) {
		//printf("! not linear: step %.12g but at %u: %.12g -> %.12g\n",
		//       dd, imax, v[imax-1], v[imax]);
		return;
	}
	ng = n;
	gi = di;
	gf = df;
	kind = _LIN; // success !
}

void CGrid::try_log(void)
{
	if (kind!=_VEC) return;
}

void CGrid::set_vec(void)
{
	if (kind!=_VEC) AsVec(&v);
	kind = _VEC;
}

void CGrid::erase (const uint begin, const uint end)
{
	GridTyp oldtype;
	set_vec();
	v.erase(v.begin()+begin, v.begin()+end);
	if (oldtype==_LIN) 
		try_lin();
	else if (oldtype==_LOG)
		try_log();
}

CGrid CGrid::Ask(const string& vnam, const uint n_in)
{
	static uint nd=0, mod1;
	static double dv=0, di=0, dd=0, df=0, gq, val;
	static int mod = -1;
	string quest, line;

	v.resize(0);
	cout << "Create grid " << vnam << ":\n";
	mod = iask_lim("Option (0:const 1:lin 2:log 3:vector)", 
		       mod, CLimits(0,3));
	switch(mod) {
	case 0:
		kind = _LIN;
		ng = iask("Number of entries", nd);
		if(!ng) break;
		nd = ng;
		gf = gi = dv = dask("Value", dv);
		return *this;
	case 1:
		kind = _LIN;
		mod1 = iask_lim("Automatic 0:n 1:start 2:step 3:stop", 
				mod1, CLimits(0,3));
		if (mod1!=0) {
			ng = iask("Number of entries", nd);
			if(!ng) break;
			nd = ng;
		}
		if (mod1!=1) gi = di = dask("Grid from", di);
		if (mod1!=2) {
			dd = dask("Step", dd);
			if (mod1==0 && dd<=0) {
				printf("! automatic n requires step>0\n");
                                // may05: forbid step<0
				break;
			}
		}
		if (mod1!=3) gf = df = dask("Up to", df);

		if        (mod1==0) {
			if ((gq=(gf-gi)/dd)<0) {
				printf("! bad sign\n");
				break;
			}
			ng = (uint) gq + 1;
		} else if (mod1==1) {
			gi = gf - (ng-1)*dd;
		} else if (mod1==2) {
		} else if (mod1==3) {
			gf = gi + (ng-1)*dd;
		}
		printf("DEBUG %u %g %g\n", ng, gi, gf);
		return *this;
	case 2:
		printf("log grid nyi\n");
		break;
	case 3:
		kind = _VEC;
		printf("Enter values:\n");
		ng = 0;
		while(1) {
			quest = "[" + strg(ng) + "] ";
			line = sask(quest.c_str());
			if (sscanf(line.c_str(), "%lg", &val)!=1) break;
			v.push_back(val);
			ng++;
		}
	default:
		break;
	}
	kind = _NOGRID;
	return *this;
}

void CGrid::clear(void)
{
	v.resize(0);
	kind = _NOGRID;
}

void CGrid::push_back(const double val)
{
	switch (kind) {
	case _NOGRID:
		kind = _LIN;
		ng = 0;
		push_back(val);
		break;
	case _LIN:
		if (ng==0) {
			gf = gi = val;
			ng = 1;
		} else if (ng==1) {
			gf = val;
			ng++;
		} else if (eq(val,gf+(gf-gi)/(ng-1),fabs(gf-gi))) {
			ng++;
			gf = val;
		} else {
			set_vec();
			v.push_back(val);
		}
		break;
	case _VEC:
		v.push_back(val);
		break;
	default:
		cout << "PROG ERR grid(invalid)::push_back\n";
	}
	// printf("DEBUG %12g -> %4u %4u %12g %12g %s\n",
	//       val, ng, size(), gi, gd, kind_str().c_str());
}

uint CGrid::size(void) const
{
	switch (kind) {
	case _NOGRID:
		return 0;
	case _LIN: case _LOG:
		return ng;
	case _VEC:
		return v.size();
	}
	return 0; // just to prevent warning
}

double CGrid::operator[](const uint i) const
{
	switch (kind) {
	case _NOGRID:
		return 0;
	case _LIN: 
		return gi + ((ng>1) ? i*(gf-gi)/(ng-1) : 0);
	case _LOG: 
		return gi * ((ng>1) ? exp(i*log(gf/gi)/(ng-1)) : 1);
	case _VEC:
		return v[i];
	default:
		return 0;
	}	
	return 0; // just to prevent warning
}

uint CGrid::AsVec(vector<double> *vout) const
{
	switch (kind) {
	case _NOGRID:
		cout << "PROG ERR grid(nogrid)::AsVec\n";
		return 0;
	case _LIN:
		vout->resize(ng);
		for(uint i=0; i<ng; i++)
			(*vout)[i] = (*this)[i];
		return ng;
	case _VEC:
		*vout = v;
		return v.size();
	default:
		cout << "PROG ERR grid(invalid)::AsVec\n";
		return 0;
	}
	return 0; // just to prevent warning
}

uint CGrid::AsVec(double *vout, size_t nout) const
{
	vector<double> vv;
	uint n;
	n = AsVec(&vv);
	if (nout<n) {
		cout << "PROG ERR AsVec nout<n\n";
		return 0;
	}
	for (uint i=0; i<n; ++i)
		vout[i] = vv[i];
	return n;
}

CRange CGrid::range(const bool logflag) const
{
	double di, df;
	switch (kind) {
	case _NOGRID:
		cout << "PROG ERR grid(nogrid)::range\n";
		return CRange();
	case _LIN:
		if (logflag&&gi<=0) printf("grid::range unfertig\n");
		return CRange(gi, gf);
	case _VEC:
		if (v.size()<=0) return CRange();
		di = df = v[0];
		for (uint i=1; i<v.size(); i++) {
			di = min (di, v[i]);
			df = max (df, v[i]);
		}
		if (logflag) {
			if (df<0) return CRange();
			di = df;
			for (uint i=1; i<v.size(); i++)
				if (v[i]>0 && v[i]<di) di = v[i];
		}
		return CRange(di, df);
	default:
		cout << "PROG ERR grid(invalid)::range\n";
		return CRange();
	}
	return CRange(); // just to prevent warning
}

int CGrid::lin(double* rd, double* ri, double* rf)
{
	if (kind!=_LIN) return 0;
	if (rd) *rd = (ng>1) ? (gf-gi)/(ng-1) : 0;
	if (ri) *ri = gi;
	if (rf) *rf = gf;
	return 1;
}

string CGrid::kind_str(void) const
{
	switch (kind) {
	case _NOGRID:
		return "???";
	case _LIN: 
		return "lin";
	case _LOG:
		return "log";
	case _VEC:
		return "vec";
	}
	return "!!!"; // just to prevent warning
}

string CGrid::str(void) const
{
	char tmp[80];
	CRange R;
	switch (kind) {
	case _NOGRID:
		return "nogrid";
	case _LIN:
		if (ng==0)
			return "ERR(lin empty)";
		else if (ng==1)
			sprintf(tmp, "%12.6g(=)%12.6g", gi, gf);
		else
			sprintf(tmp, "%12.6g(%12.6g)%12.6g", 
				gi, (gf-gi)/(ng-1), gf);
		return string(tmp);
	case _LOG:
		if (ng==0)
			return "ERR(log empty)";
		else if (ng==1)
			sprintf(tmp, "%12.6g[=]%12.6g", gi, gf);
		else
			sprintf(tmp, "%12.6g[%8u]%12.6g", 
				gi, ng, gf);
		return string(tmp);
	case _VEC:
		R = range();
		sprintf(tmp, "%15.9g..%15.9g", R.inf, R.sup);
		return string(tmp);
	}
	return "!!!"; // just to prevent warning
}

int CGrid::pos(double c) const // returns values from -1 to n
{
	int ii;
	uint i, n;
	switch (kind) {
	case _LIN:
		if (ng<=1) {
			printf("apply grid::pos() with n=1 ??\n");
			return 0;
		}
		if (gi>=gf) {
			printf("grid::pos() lin grid not sorted:"
			       "gi=%g, gf=%g\n", gi, gf);
			return 0;
		}
		ii = (int) floor((c-gi)*(ng-1)/(gf-gi) + .5);
		return min((int)ng,max(-1,ii));
	case _LOG:
		printf("PROG NOT READY grid::pos() log\n");
		return -1;
	case _VEC:
		n=v.size();
		if (n<=1) {
			printf("apply grid::pos() with n=1 ??\n");
			return 0;
		}
		// assume v is sorted
		if (c < v[0] - (v[1]-v[0])/2) return -1;
		for(i=0; i<n-1; i++)
			if (c < (v[i]+v[i+1])/2) return i;
		if (c < v[n-1] + (v[n-1]-v[n-2])/2) return n-1;
		return n;
	case _NOGRID: default:
		printf("PROG ERROR grid::pos() nogrid\n");
		return -1;		
	}
	return -2;
}

uint CGrid::pos_force(double c) const 
{
	return min((int)size()-1, (int)max(0, pos(c)));
}

bool CGrid::compatible(const CGrid& G2, string *err) const
{
	if (kind==_NOGRID) {
		if (err) *err = "grid 1 undefined";
		return 0;
	}
	if (G2.kind==_NOGRID) {
		if (err) *err = "grid 2 undefined";
		return 0;
	}
	uint n = size();
	if (!n) {
		if (err) *err = "grid 1 empty";
		return 0;
	}
	if (n!=G2.size()) {
		if (err) *err = "grids have different lengths";
		return 0;
	}
	if (kind==_LOG || G2.kind==_LOG) {
		if (err) *err = "compatibility test for _LOG grid missing";
		return 0;
	}
	if (kind==_LIN ^ G2.kind==_LIN) {
		if (err) *err = "only one of two grids is known to be linear";
		return 0;
	}
	double tol = (sup() - inf()) * grid_tol;
	if (kind==_VEC) {
		for (uint i=0; i<n; ++i)
			if (fabs((*this)[i] - G2[i]) > tol) {
				if (err) *err = "grids differ at pos "+strg(i) 
						 + ": g1=" + strg((*this)[i]) 
						 + ", g2=" + strg(G2[i]);
				return 0;
			}
		return 1;
	}
	if (kind==_LIN) {
		if (fabs(inf()-G2.inf()) > tol) {
			if (err) *err = "lower bounds differ"\
					   ": g1=" + strg(inf()) 
					 + ", g2=" + strg(G2.inf());
			return 0;
		}
		if (fabs(sup()-G2.sup()) > tol) {
			if (err) *err = "upper bounds differ"\
					   ": g1=" + strg(sup()) 
					 + ", g2=" + strg(G2.sup());
			return 0;
		}
		return 1;
	}
	printf("!!! PROG ERR Grid::compatible unexpected case\n");
	return 0;
}

bool CGrid::eq(const double r1, const double r2, const double step) const
{
	return (fabs(r1-r2) < fabs(step)*grid_tol);
}

const double CGrid::grid_tol=1e-6;
