//**************************************************************************//
//* FRIDA: fast reliable interactive data analysis                         *//
//* manip.cpp: manipulate data (select, average, rearrange, ...)           *//
//* (C) Joachim Wuttke 1990-, v2(C++) 2001-                                *//
//* http://frida.sourceforge.net                                           *//
//**************************************************************************//

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <iostream>
#include <string>
#include <vector>
#include <algorithm>
#include <gsl/gsl_sort.h>

#include "olm.h"
#include "asi.h"
#include "manip.h"
#include "edif.h" // wg. ListZ()
#include "xax.h"

using namespace std;

// local routines:
namespace NManip {
    void ScaExeMerge( COld *fin, COld *fout, const CList& BP );
    void ScaExeAvge( COld *fin, COld *fout, const CList& BP );
    void ScaExeBreak( COld *fin, const COld& ftmp, const CList& BP );
    void ScaRemoveConstantZ( COld *f );
    int PtsExeRetEqui( const vector<double> x, uint nblock, double tol, 
                       IndexSet *Sel);
}


//***************************************************************************//
//* manipulations on data points                                            *//
//***************************************************************************//

void NManip::PtsDelete(void)
{
    cout << "Points delete options:\n"
        "  (1) delete by number         (2) retain by number\n"
        "  (3) delete if x in range     (4) retain if x in range\n"
        "  (5) delete if y in range     (6) retain if y in range\n";
    static int mOpt = -1;
    mOpt = iask_lim("Option", mOpt, CLimits(0,6));
    if (!mOpt) return;

    bool seldel = mOpt % 2;
    int  mCrit  = (mOpt+1) / 2;
	
    static CList LSel;
    static CRange RSel;
    switch (mCrit) {
    case 1:
        LSel.Ask("Which points", LSel);
        break;
    case 2:
    case 3:
        RSel.Ask("Which range", RSel);
        break;
    default:
        cout << "! invalid mCrit\n";
        return;
    }

	
    COld *fin, fout;
    CScan *Sin, Sout;
    bool crit;
	
    NOlm::IterateD fiter;
    NOlm::SelNew();
    while((fin=fiter())) {
        fout = *fin;
        fout.lDoc.push_back("delete points ...");
        fout.VS.clear();

        for (uint j=0; j<fin->VS.size(); j++) {
            Sin = &(fin->VS[j]);
            Sout = *Sin;
            Sout.x.clear();
            Sout.y.clear();
            for (uint i=0; i<Sin->size(); i++) {
                switch (mCrit) {
                case 1:
                    crit = LSel.contained(i);
                    break;
                case 2:
                    crit = RSel.contained(Sin->x[i]);
                    break;
                case 3:
                    crit = RSel.contained(Sin->y[i]);
                    break;
                }
                if (seldel ^ crit)
                    Sout.push_back(Sin->x[i], Sin->y[i]);
            }
            if (Sout.size())
                fout.VS.push_back(Sout);
        }
        if (fout.VS.size())
            NOlm::OloAdd(&fout);
        else 
            cout << "no data point left\n";
    }
}

void NManip::PtsAvge(void)
{
    static string groups;
    IndexSet Groups;
    IndexSetIterator ISet;
    uint i, igi, igf, ng;
    double xg, yg;

    groups = index_ask("Start groups at points", groups);
    if (groups=="") return;
	
    COld *fin, fout;
    CScan *Sin, Sout;

    NOlm::IterateD fiter;
    NOlm::SelNew();
    while((fin=fiter())) {
        fout = *fin;
        fout.lDoc.push_back("average points at " + groups);
        fout.VS.clear();

        for (uint j=0; j<fin->VS.size(); j++) {
            Sin = &(fin->VS[j]);
            Sout = *Sin;
            Sout.x.clear();
            Sout.y.clear();
            if (Groups.parse(groups, 0, Sin->size())) {
                printf("File %u, scan %u: invalid group\n",
                       fiter.SelNo(), j);
                return;
            }
            if (!Groups.contained(0)) {
                printf("Group must start at 0\n");
                return;
            }
            Groups.insert(Sin->size());

            ISet.reset(Groups);
            while (ISet(&igi) && ISet(&igf)) {
                // printf("DEBUG igi=%u igf=%u\n", igi, igf);
                ISet.stepback();
                xg = yg = 0;
                for (i=igi; i<igf; ++i) {
                    xg += Sin->x[i];
                    yg += Sin->y[i];
                }
                ng = igf - igi;
                Sout.push_back(xg/ng, yg/ng);
            }
            fout.VS.push_back(Sout);
        }
        NOlm::OloAdd(&fout);
    }
}

void NManip::PtsBreak(void)
{
    // determine scan length:
    COld *fin, fout;
    NOlm::IterateD fiter;
    uint np = 0;
    while((fin=fiter())) {
        np = max (np, fin->nPts());
    }
	
    // ask for break points:
    static CList PSel;
    PSel.Ask("Subscans start at entry numbers", PSel, CLimits(0,np-1));
	
    // loop over input files:
    CScan *Sin, Sout;
    fiter.Reset();
    NOlm::SelNew();
    while((fin=fiter())) {

				
        // new z, documentation:
        fout = *fin;
        fout.VS.clear();
        fout.ZCo.push_back(CCoord(fin->xco.name + "_break", 
                                  fin->xco.unit));
        fout.lDoc.push_back("break scans at " + PSel.str());
		
        // execute per scan:
        for(uint j=0; j<fin->nScan(); ++j) {

            np = fin->nPts(j);
            PSel.ProvideLims(0, np-1);
            if (!PSel.synvalid()) {
                printf("!! invalid limits\n");
                return;
            }

            Sin = &(fin->VS[j]);
			
            uint i0, ib, ii;
            ListIter piter(PSel);
            piter(&i0);
            do {
                if (!(ii = piter(&ib))) ib = np;
				
                // new scan from i0 to ib-1:
                //printf("DEB subscan from %d to %d-1\n",i0,ib);
                if (ib<i0) {
                    printf("!! invalid break %d < %d\n",
                           ib, i0);
                    return;
                }
				
                Sout.Clear();
                for (uint i=i0; i<ib; ++i)
                    Sout.push_back(Sin->x[i], Sin->y[i]);
                Sout.z = Sin->z;
                Sout.z.push_back(Sout.x[0]);
                fout.VS.push_back(Sout);
				
                i0 = ib;
            } while (ii);
        }
        NOlm::OloAdd(&fout);
    }
}

void NManip::PtsSort( void )
{
    string expr = sask("Sort points according to");
    if (expr=="") return;
    CTree T;
    if (user_xaxparse(expr.c_str(), &T)) {
        printf("! invalid expression\n");
        return;
    }

    NOlm::IterateD fiter;
    if (!NOlm::overwrite) NOlm::SelNew();
    COld *fin, fout;
    COlo *f;
    CScan *Sin, Sout;
    double xval, zval, result;
    uint k, nz, n, i;
    while((fin=fiter())) {
        k = fiter.SelNo();
        f = (COlo*) fin;

        fout = *fin;
        fout.VS.clear();
        fout.lDoc.push_back("mpo according to "+expr);

        for (uint j=0; j<fin->nScan(); j++) {
            Sin = &(fin->VS[j]);
            n = Sin->size();
            double v[n];
            for (i=0; i<n; i++)
                if (T.tree_val(&(v[i]), f, k, j, i))
                    return;
            size_t P[n];
            double xin[n];
            Sin->x.AsVec(xin, n);

            gsl_sort_index (P, v, 1, n);
                
            Sout = *Sin;
            double xout[n];
            for (i=0; i<n; ++i) {
                xout[i] = xin[P[i]];
                Sout.y[i] = Sin->y[P[i]];
            }

            Sout.x = CGrid(xout, n);
            fout.VS.push_back(Sout);
        }
        NOlm::OloAdd(&fout);
    }
}

void NManip::PtsAvgeEq(void)
{
    uint j, i, ii, n, ng;
    double xg, yg;

    COld *fin, fout;
    CScan *Sin, Sout;

    NOlm::IterateD fiter;
    NOlm::SelNew();
    while((fin=fiter())) {
        fout = *fin;
        fout.lDoc.push_back("average points when x equal");
        fout.VS.clear();

        for (j=0; j<fin->VS.size(); j++) {
            Sin = &(fin->VS[j]);
            n = Sin->size();
            Sout = *Sin;
            Sout.x.clear();
            Sout.y.clear();

            for (i=0; i<n; i=ii) {
                xg = Sin->x[i];
                yg = Sin->y[i];
                ng = 1;
                for (ii=i+1; ii<n; ++ii) {
                    if (Sin->x[ii] != xg) break;
                    yg += Sin->y[ii];
                    ++ng;
                }
                Sout.push_back(xg, yg/ng);
            }
            fout.VS.push_back(Sout);
        }
        NOlm::OloAdd(&fout);
    }
}

void NManip::PtsRetainEqui(void)
{
    COld *fin, fout;
    CScan *Sin, Sout;
    vector<double> x;
    uint n, i, ii, offs;
    static uint nblock=1;
    double step, tol=1e-6;

    nblock = iask("Block size", nblock);
    if (nblock==0) return;

    NOlm::IterateD fiter;
    NOlm::SelNew();
    while((fin=fiter())) {
        fout = *fin;
        fout.lDoc.push_back("delete points unless x is equidistant...");
        fout.VS.clear();

        for (uint j=0; j<fin->VS.size(); j++) {
            Sin = &(fin->VS[j]);
            Sout = *Sin;

            Sin->x.AsVec(&x);

            n = x.size();
            VTEST ((n<2), ("!!! not enough data points\n"));

            // determine step :
            step = x[1] - x[0];

            // retain blocks:
            offs=0;
            for (i=0; i<=n-nblock; ) {
                for (ii=i+1; ii<i+nblock; ii++) {
                    if (fabs(x[ii]-x[ii-1]-step) > tol) {
                        printf("unselect %u to %u\n", 
                               i, ii-1);
                        Sout.erase(offs+i,offs+ii);
                        offs -= ii-i;
                        break;
                    } 
                }
                i = ii;
            }
            if (i<n) printf("overhang %u to %u\n", i, n-1);
            Sout.erase(offs+i, offs+n);
			
            if (Sout.size())
                fout.VS.push_back(Sout);
        }
        if (fout.VS.size())
            NOlm::OloAdd(&fout);
        else 
            cout << "no data point left\n";
    }
}

void NManip::PtsGroupWeighted( void )
// for use in PISA jan07
// input: each scan's y tuple contains: xg, wgt, y0, .., ym
//    where xg is a coordinate according to which the y0 .. ym will be grouped
//    where wgt is the weight of the tuple
{
    // cout << "warning:\n"
    //    "will fail if input point ought to be distributed"
    //    " in more than 2 groups\n";
    static int ng=40;
    ng = iask( "number of groups", ng );
    if( ng<2 ) return;

    NOlm::IterateD fiter;
    if (!NOlm::overwrite) NOlm::SelNew();
    COld *fin, fout, fout2;
    COlo *f;
    CScan *Sin, Sout, Sout2;
    double wgttot, wgtgrp, wgtacc, frac;
    uint k, nz, n, ny, i, iy, igrp;
    vector<double> xtmp, wtmp;

    while((fin=fiter())) {
        k = fiter.SelNo();
        nz = fin->ZCo.size();
        f = (COlo*) fin;

        ny = fin->VS[0].size()-2;
        if( (int)ny<1 ){
            printf( "ny<1\n" );
            return;
        }

        fout = *fin;
        fout.VS.clear();

        fout.xco = CCoord( "group", "" );
        fout.ZCo.pop_back();
        if( ny>1 ) fout.ZCo.push_back( CCoord( "iy", "" ) );

        fout2 = fout;
        fout2.yco.name = "stdv["+fout.yco.name+"]";
        fout.lDoc.push_back("mpgw ng="+strg(ng));
        fout2.lDoc.push_back("error bars of mpgw ng="+strg(ng));

        Sout.Clear();
        Sout2.Clear();

        vector<double> xgrp(ng);
        vector< vector<double> > ytmp, ygrp, ygrp2;
        ytmp.resize( ny );
        ygrp.resize( ny );
        ygrp2.resize( ny );
        for( iy=0; iy<ny; ++iy ){
            ytmp [iy].clear();
            ygrp [iy].resize(ng);
            ygrp2[iy].resize(ng);
        }

        for (uint j=0; j<fin->nScan(); j++) {

            Sin = &(fin->VS[j]);
            if( Sin->size()-2 != ny ){
                printf( "ny not constant\n" );
                return;
            }
            Sout.z = Sin->z;
            Sout.z.pop_back();
            Sout2.z = Sin->z;
            Sout2.z.pop_back();
            xtmp.push_back( Sin->y[0] );
            wtmp.push_back( Sin->y[1] );
            for( iy=0; iy<ny; ++iy )
                ytmp[iy].push_back( Sin->y[2+iy] );

            if ( j+1==fin->VS.size() || 
                 nz>=2 && fin->VS[j+1].z[nz-2]!=Sin->z[nz-2] ) {
                // reached end-of-file or end-of-relevant-z-group
                // sort, group, save, possibly start new scan
                n = xtmp.size();
                size_t P[n];
                double xin[n];
                for (i=0; i<n; ++i) { xin[i] = xtmp[i]; }
                gsl_sort_index( P, xin, 1, n );
                vector<double> xsor(n), wsor(n);
                vector< vector<double> > ysor(ny);
                for( iy=0; iy<ny; ++iy ){
                    ysor[iy].resize(n);
                }
                for (i=0; i<n; ++i) {
                    xsor[i] = xtmp[P[i]];
                    wsor[i] = wtmp[P[i]];
                    for( iy=0; iy<ny; ++iy )
                        ysor[iy][i] = ytmp[iy][P[i]];
                }
                for( igrp=0; igrp<ng; ++igrp ){
                    xgrp[igrp] = 0;
                    for( iy=0; iy<ny; ++iy ){
                        ygrp[iy][igrp] = 0;
                        ygrp2[iy][igrp] = 0;
                    }
                }
                wgttot = 0;
                for( i=0; i<n; ++i ){ wgttot += wsor[i]; }
                wgtgrp = wgttot/ng;
                wgtacc = 0;
                for( i=0, igrp=0; i<n; ++i ){
                    if( wgtacc+wsor[i]>(igrp+1)*wgtgrp ){
                        frac = (igrp+1)*wgtgrp-wgtacc;
                        xgrp[igrp] += frac * xsor[i] / wgtgrp;
                        for( iy=0; iy<ny; ++iy ){
                            ygrp[iy][igrp] += frac * ysor[iy][i] / wgtgrp;
                            ygrp2[iy][igrp] += frac* SQR(ysor[iy][i]) /wgtgrp;
                        }
                        ++igrp;
                        if( igrp>=ng ) break;
                        frac = wgtacc+wsor[i]-igrp*wgtgrp;
                        xgrp[igrp] += frac * xsor[i] / wgtgrp;
                        for( iy=0; iy<ny; ++iy ){
                            ygrp[iy][igrp] += frac * ysor[iy][i] / wgtgrp;
                            ygrp2[iy][igrp] += frac* SQR(ysor[iy][i]) /wgtgrp;
                        }
                    } else {
                        xgrp[igrp] += wsor[i] * xsor[i] / wgtgrp;
                        for( iy=0; iy<ny; ++iy ){
                            ygrp[iy][igrp] += wsor[i] * ysor[iy][i] / wgtgrp;
                            ygrp2[iy][igrp] += wsor[i]*SQR(ysor[iy][i])/wgtgrp;
                        }
                    }
                    wgtacc += wsor[i]; 
                }
                for( iy=0; iy<ny; ++iy ){
                    if( ny>1 ){
                        Sout.z.push_back( iy );
                        Sout2.z.push_back( iy );
                    }
                    for( igrp=0; igrp<ng; ++igrp ){
                        Sout.push_back( xgrp[igrp], ygrp[iy][igrp] );
                        Sout2.push_back( xgrp[igrp], 
                            sqrt( ygrp2[iy][igrp] - SQR(ygrp[iy][igrp]) ) );
                    }
                    fout.VS.push_back(Sout);
                    Sout.Clear();
                    Sout.z = Sin->z;
                    Sout.z.pop_back();
                    fout2.VS.push_back(Sout2);
                    Sout2.Clear();
                    Sout2.z = Sin->z;
                    Sout2.z.pop_back();
                }
                xtmp.clear();
                wtmp.clear();
                for( iy=0; iy<ny; ++iy )
                    ytmp[iy].clear();
            }
        }
        NOlm::OloAdd(&fout);
        NOlm::OloAdd(&fout2);
    }
}


//***************************************************************************//
//* manipulations on scans                                                  *//
//***************************************************************************//

void NManip::ScaRemoveConstantZ( COld *f )
{
    uint nz = f->nZ();
    for( int iz=nz-1; iz>=0; --iz ){
        uint nj = f->VS.size();
        if( nj<1 ){
            printf( "WARNING: file contains no scan\n" );
            return;
        }
        double z = f->VS[0].z[iz];
        bool z_is_const = true;
        for( uint j=1; j<nj; ++j ){
            if( f->VS[j].z[iz] != z ){
                z_is_const = false;
            }
        }
        if( z_is_const ){
            printf( " constant z%i becomes r%i\n", iz, f->RPar.size() );
            f->RPar.push_back( CParam( f->ZCo[iz], z ) );
            f->ZCo.erase( f->ZCo.begin()+iz );
            for( uint j=0; j<f->VS.size(); ++j )
                f->VS[j].z.erase( f->VS[j].z.begin()+iz );
        }
    }
}

void NManip::ScaDelete(void)
{
    if (NOlm::FJSelAsk("Delete which scans")) return;

    COld *fin, fout;
    CList JSel;
    NOlm::IterateD fiter;
    NOlm::SelNew();
    while((fin=fiter())) {
        fout = *fin;

        NOlm::FJSelGet(fiter.SelNo(), &JSel);
        fout.lDoc.push_back("delete scans " + JSel.str());
        JSel.SortReverse();
        IListV<CScan> jiter(&(fin->VS), &JSel);
        while (jiter()) {
            fout.VS.erase(fout.VS.begin()+jiter.SelNo());
        }

        ScaRemoveConstantZ( &fout );
        NOlm::OloAdd(&fout);
    }
}

void NManip::ScaAvge(void)
{
    if (NOlm::FJSelAsk("Start groups at scans")) return;
    static bool skip_zero = false;
    // skip_zero = bask( "Does 0 mean \"ignore this scan\"", skip_zero );

    COld *fin, fout;
    CScan Sout;
    CList JSel;
    vector<double> y;
    uint ji, jf, j, i, n, mj, nz;
    double ym;
    string err;
    NOlm::IterateD fiter;
    NOlm::SelNew();
    while((fin=fiter())) {
        fout = *fin;

        NOlm::FJSelGet(fiter.SelNo(), &JSel);
        if (JSel.V.size()<1 || JSel.V[0]!=0) {
            printf("scan selection must contain 0\n");
            return;
        }
        fout.lDoc.push_back("average scans " + JSel.str());

        fout.VS.clear();
        nz = fin->nZ();
        for(uint iv=0; iv<JSel.n(); iv++) {
            ji = JSel.V[iv];
            jf = ( iv<JSel.n()-1 ? JSel.V[iv+1] : fin->nScan() );
            Sout = fin->VS[ji];
            n = fin->VS[ji].size();
            for (j=ji+1; j<jf; j++) {
                if (!fin->VS[j].x.compatible(
                        fin->VS[ji].x, &err)) {
                    printf("! grids %d and %d differ: %s\n",
                           j, ji, err.c_str());
                    return;
                }
            }
            y.resize(n);
            for (i=0; i<n; i++){
                mj = 0;
                ym = 0;
                for (j=ji; j<jf; j++) {
                    if( skip_zero && fin->VS[j].y[i]==0 ) continue;
                    ym += fin->VS[j].y[i];
                    ++mj;
                }
                y[i] = mj==0 ? 0 : ym/mj;
            }
            Sout.y = y;
            for( uint iz=0; iz<nz; ++iz ){
                uint mz = 0;
                double z = 0;
                for (j=ji; j<jf; j++) {
                    z += fin->VS[j].z[iz];
                    ++mz;
                }
                Sout.z[iz] = z/mz;
            }
            fout.VS.push_back(Sout);
        }
        ScaRemoveConstantZ( &fout );
        NOlm::OloAdd(&fout);
    }
}

void NManip::ScaSpawn(void)
{
    static int njj = 2;
    int njjIn = iask("output will have how many copies of input scan", njj);
    if (njjIn<1) return;
    njj = njjIn;

    COld *fin, fout;
    CScan Sout;
    uint jj, j;
    NOlm::IterateD fiter;
    NOlm::SelNew();
    while((fin=fiter())) {
        fout = *fin;

        fout.ZCo.push_back(CCoord("no-in-spawn", ""));
        fout.VS.clear();

        for (jj=0; jj<njj; ++jj) {
            for (j=0; j<fin->VS.size(); j++) {
                Sout = fin->VS[j];
                Sout.z.push_back(jj);
                fout.VS.push_back(Sout);
            }
        }
        NOlm::OloAdd(&fout);
    }
}

void NManip::ScaGridSimplify(void)
{
    COld *fin;
    NOlm::IterateD fiter;
    while((fin=fiter()))
        for (uint j=0; j<fin->nScan(); j++)
            fin->VS[j].x.simplify();
}

void NManip::ScaExch(void)
{
    COld *fin, fout;
    CScan Sout;
    vector<double> zcommon, xcommon;
    uint ji, jf, j, i, ii, izco;
    CCoord xcoin, zcoin;
    string doc;

    NOlm::IterateD fiter;
    NOlm::SelNew();
    while((fin=fiter())) {
        fout = *fin;
        fout.VS.clear();

        xcoin = fin->xco;
        izco = fin->ZCo.size() - 1;
        VTEST ((izco==(uint)-1), ("! no input z coordinate\n"));
        zcoin = fin->ZCo[izco];

        doc = "exchanging x=" + xcoin.str() + " against z" + strg(izco)
            + "=" + zcoin.str();
        cout << doc+"\n";
	
        fout.lDoc.push_back(doc);
        fout.xco = zcoin;
        fout.ZCo[izco] = xcoin;
		
        VTEST ((fin->VS.size()<=0), ("no scans in file\n"));

        for (ji=0; ji<fin->VS.size(); ) {

            zcommon.clear();
            for (i=0; i<izco; ++i)
                zcommon.push_back(fin->VS[ji].z[i]);

            for (jf=ji+1; jf<fin->VS.size(); jf++)
                for (i=0; i<izco; ++i)
                    if (fin->VS[jf].z[i] != zcommon[i])
                        goto end_group;
        end_group:

            // build common grid:
            xcommon.clear();
            for (j=ji; j<jf; ++j)
                for (i=0; i<fin->VS[j].size(); ++i)
                    xcommon.push_back(fin->VS[j].x[i]);
            sort(xcommon.begin(), xcommon.end());
            mystd::unique(&xcommon, 1e-180, 1e-20);

            printf("group of scans %u .. %u has %u different x\n", 
                   ji, jf-1, xcommon.size());

            ii = 0; // a guess
            for(i=0; i<xcommon.size(); ++i) {
                Sout.Clear();
                Sout.z = zcommon;
                Sout.z.push_back(xcommon[i]);

                for (j=ji; j<jf; ++j){
                    if( fin->VS[j].x[ii]!=xcommon[i] ){ // guess failed
                        for (ii=0; ii<fin->VS[j].size(); ++ii)
                            if( fin->VS[j].x[ii]==xcommon[i] )
                                break;
                    }
                    if( ii<fin->VS[j].size() )
                        Sout.push_back( fin->VS[j].z[izco], fin->VS[j].y[ii] );
                }
                fout.VS.push_back(Sout);
                ++ii;
            }
            ji = jf;
        }
        NOlm::OloAdd(&fout);
    }
}

void NManip::ScaSortByExpr( void )
{
    string expr = sask("Sort spectra according to");
    if (expr=="") return;
    CTree T;
    if (user_xaxparse(expr.c_str(), &T)) {
        printf("! invalid expression\n");
        return;
    }

    NOlm::IterateD fiter;
    if (!NOlm::overwrite) NOlm::SelNew();
    COld *fin, fout;
    COlo *f;
    CScan *Sin, Sout;
    double xval, zval, result;
    uint k, nj, i;
    while((fin=fiter())) {
        k = fiter.SelNo();
        f = (COlo*) fin;

        fout = *fin;
        fout.VS.clear();
        fout.lDoc.push_back("mso according to "+expr);

        nj = fin->nScan();
        double v[nj];
        size_t P[nj];
        for (uint j=0; j<nj; j++) {
            Sin = &(fin->VS[j]);
            if (T.tree_val(&(v[j]), f, k, j))
                return;
        }
        
        gsl_sort_index (P, v, 1, nj);
                
        for (uint j=0; j<nj; j++) {
            fout.VS.push_back( fin->VS[P[j]] );
        }
        NOlm::OloAdd(&fout);
    }
}

void NManip::ScaSortByZ( void )
{
    COld *fin, fout;
    NOlm::IterateD fiter;
    NOlm::SelNew();
    while((fin=fiter())) {
        fout = *fin;
        sort( fout.VS.begin(), fout.VS.end(), CScanSortByZ );
        NOlm::OloAdd(&fout);
    }
}

void NManip::ZExchange( void )
{
    int nzmax = 0;
    int nzmin;
    char mod;
    string com;
    int num1;
    int num2;
    COld *fin, fout;
    NOlm::IterateD fiter;
    while( (fin=fiter()) ){
        nzmax = max( (int)nzmax, (int)fin->ZCo.size() );
    }
    if       ( nzmax<=1 ){
        printf( "nothing to exchange\n" );
        return;
    } else if( nzmax==2 ){
        mod = 'r';
        printf( "exchanging two z coordinates\n" );
        com = "exchanged two z coordinates";
        num1 = 1;
        nzmin = 2;
    } else {
        string sel;
        sel = wask( "rotate(r,rr,...), revert(m), exchange(<number>)" );
        if      ( sel[0]=='r' ){
            mod = 'r';
            num1 = 1;
            while( sel.length() > num1 ){
                if( sel[num1++]!='r' ){
                    printf( "invalid selection\n" );
                    return;
                }
            }
            com = "rotated z coordinates by " + strg(num1);
            nzmin = 2;
        } else if( sel=="m" ){
            mod = 'm';
            com = "reverted z coordinates";
            nzmin = 2;
        } else {
            if( sscanf(sel.c_str(), "%d", &num1)!=1 ){
                printf( "invalid selection\n" );
                return;
            }
            if( num1>=nzmax ){
                printf( "no more than %i z coordinates\n", nzmax );
                return;
            }
            num2 = iask( "exchange with coordinate" );
            if( num2>=nzmax ){
                printf( "no more than %i z coordinates\n", nzmax );
                return;
            } else if( num1==num1 ){
                printf( "nothing to exchange" );
                return;
            }
            mod = 'x';
            com = "exchanged z coordinates "+strg(num1)+" and "+strg(num2);
            nzmin = max( num1, num2 )+1;
        }
    }
                
    NOlm::SelNew();
    fiter.Reset();
    while((fin=fiter())) {
        fout = *fin;
        fout.lDoc.push_back( com );
        int nz = fout.ZCo.size();
        if      ( nz<nzmin ){
            printf( "WARNING nz<nzmin\n" );
        } else if( mod=='r' ){
            for( int irot=0; irot<num1; ++irot ){
                for( int iz=1; iz<nz; ++iz ){
                    fout.ZCo[iz] = fin->ZCo[iz-1];
                }
                fout.ZCo[0] = fin->ZCo[nz-1];
            }
        } else if( mod=='m' ){
            for( int iz=0; iz<nz; ++iz ){
                fout.ZCo[iz] = fin->ZCo[nz-1-iz];
            }
        } else { // mod=='x'
            fout.ZCo[num1] = fin->ZCo[num2];
            fout.ZCo[num2] = fin->ZCo[num1];
        }

        for (uint j=0; j<fin->VS.size(); j++) {
            if      ( nz<nzmin ){
                // do nothing
            } else if( mod=='r' ){
                for( int irot=0; irot<num1; ++irot ){
                    for( int iz=1; iz<nz; ++iz ){
                        fout.VS[j].z[iz] = fin->VS[j].z[iz-1];
                    }
                    fout.VS[j].z[0] = fin->VS[j].z[nz-1];
                }
            } else if( mod=='m' ){
                for( int iz=0; iz<nz; ++iz ){
                    fout.VS[j].z[iz] = fin->VS[j].z[nz-1-iz];
                }
            } else  { // mod=='x'
                fout.VS[j].z[num1] = fin->VS[j].z[num2];
                fout.VS[j].z[num2] = fin->VS[j].z[num1];
            }
        }
        NOlm::OloAdd(&fout);
    }

}


void NManip::ZDelete(void)
{
    string sel;
    NEdif::ListZ();
    sel = wask("Delete z coordinates (-=quit)");
    if (sel=="-" || sel=="q") return;
	
    COld *fin, fout;

    NOlm::IterateD fiter;
    IndexSet ISel;
    IndexSetIterator II;
    uint iz;

    NOlm::SelNew();
    while((fin=fiter())) {
        fout = *fin;
        fout.lDoc.push_back("delete z coordinates: "+sel);

        VTEST ((ISel.parse(sel, 0, fin->ZCo.size())), \
               ("! invalid selection for file %u\n", fiter.SelNo()));

        II.setback(ISel);
        while (II(&iz)) {
            fout.ZCo.erase(fout.ZCo.begin()+iz);
            for (uint j=0; j<fin->VS.size(); j++) 
                fout.VS[j].z.erase(fout.VS[j].z.begin()+iz);
        }

        NOlm::OloAdd(&fout);
    }
}

void NManip::ScaZ(void)
{
    COld *fin;
    NOlm::IterateD fiter;

    vector<CCoord> TabCo;
    vector<char>   TabSt; // status
    CList TabSel, TestList;
    ListIter tsiter;
    string   breaktext="";
    bool     breakcoord;
    uint     i, ii, j, jj, maxscan;
    char action;

    /** make coord table **/
    fiter.Reset();
    while((fin=fiter())) {
        for (i=0; i<fin->ZCo.size(); ++i) {
            for (ii=0; ii<TabCo.size(); ++ii)
                if (fin->ZCo[i]==TabCo[ii]) goto oldzco;
            TabCo.push_back(fin->ZCo[i]);
            TabSt.push_back('d');
        oldzco: ;
        }
    }
	
    /** find maxscan **/
    maxscan = 0;
    fiter.Reset();
    while((fin=fiter()))
        maxscan = max(maxscan, fin->nScan());

    /** set status of z coordinates **/
    while(1) {

        /** display table **/
        printf("Status of z coordinates:"
               " d(elete a(verage b(reakpoint\n");
        breakcoord = 0;
        for (ii=0; ii<TabCo.size(); ++ii) {
            printf(" %3d [%c] %s\n", 
                   ii, TabSt[ii], TabCo[ii].str().c_str());
            if (TabSt[ii]=='b') breakcoord = 1;
        }
        if (breakcoord) {
            breaktext = "";
            printf("Breakpoints according to coordinates\n");
        } else if (breaktext!="") {
            printf("Breakpoints: %s\n", breaktext.c_str());
        } else {
            printf("Default breakpoints: 0 for A/M, * for B\n");
        }

        /** user action **/
    ask_action:
        action = cask("Action", 'h');
        switch (action) {
        case 'q':
            return;
        case 'd': case 'a': case 'b':
            TabSel.Ask("Set status of coordinate no.",
                       CLimits(0, TabCo.size()-1));
            tsiter.Reset(TabSel);
            while(tsiter(&ii))
                TabSt[ii] = action;
            break;
        case 's':
            breaktext = wask("Enter list of breakpoints");
            if (breaktext=="") break;
            TestList = CList(breaktext, CLimits(0, maxscan-1));
            if (!TestList.semvalid()) {
                printf("! invalid list\n");
                break;
            } else if (!TestList.contained(0)) {
                printf("! breakpoints must include 0\n");
                break;
            }
            for (ii=0; ii<TabCo.size(); ++ii)
                if (TabSt[ii]=='b') TabSt[ii]='d';
            break;
        case 'A': case 'M': case 'B':
            goto execute;
        default:
            if (action!='h') mystd::bell();
            printf("possible actions:\n"
                   "   d a b    change status of coordinates\n"
                   "   s        set list of breakpoints\n"
                   "   A M      average / merge scans\n"
                   "   B        break one file into several\n");
            goto ask_action;
        } 
    }

    /** call execute routines **/
 execute:

    COld fout;
    string docline;

    fiter.Reset();
    NOlm::SelNew();
    while ((fin=fiter())) {

        fout = *fin;
        fout.VS.clear();
        fout.ZCo.clear();

        if (fin->nScan()<=0) {
            printf("!! Empty file\n");
            return;
        }

        /** get coordinate status **/
        vector<char> CoSt(fin->ZCo.size());
        for (i=0; i<fin->ZCo.size(); ++i) {
            for (ii=0; ii<TabCo.size(); ++ii)
                if (fin->ZCo[i]==TabCo[ii]) goto entryfound;
            printf("!!! PROG ERR: coord not found\n");
            return;
        entryfound:
            CoSt[i] = TabSt[ii];
        }

        /** get break points **/
        CList BP;
        if (breakcoord) {
            vector<double> Z(fin->ZCo.size());
            Z = fin->VS[0].z;
            BP = CList((uint) 0);
            for (j=1; j<fin->nScan(); ++j) {
                for (i=0; i<fin->ZCo.size(); ++i) {
                    if (CoSt[i]=='b' && 
                        fin->VS[j].z[i]!=Z[i]) {
                        Z[i] = fin->VS[j].z[i];
                        BP += CList(j);
                    }
                }
            }
        } else if (breaktext!="") {
            BP = CList(breaktext, CLimits(0, fin->nScan()-1));
            if (!BP.semvalid()) {
                printf("! list of breakpoints invalid\n");
                return;
            }
        } else if (action=='B') {
            BP = CList (0, fin->nScan()-1);
        } else {
            BP = CList ((uint) 0);
        }

        /** generate z coordinates **/
        for (i=0; i<fin->ZCo.size(); ++i) {
            if        (CoSt[i]!='d') {
                fout.ZCo.push_back(fin->ZCo[i]);
            } 
        }
		
        /** docline **/
        switch (action) {
        case 'A':
            docline = "average scans";
            break;
        case 'M':
            docline = "merge scans";
            break;
        case 'B':
            docline = "break file into several";
            break;
        default:
            printf("!!! PROG ERROR invalid action\n");
            return;
        }
        docline += " breakpoints [" + BP.str() + "]";
        fout.lDoc.push_back(docline);

        /** new scans, set z **/
        for (j=0; j<fin->nScan(); ++j) {
            if (!BP.contained(j)) continue;
            CScan S;
            for (i=0; i<fin->ZCo.size(); ++i) {
                if        (CoSt[i]=='b') {
                    S.z.push_back(fin->VS[j].z[i]);
                } else if (CoSt[i]=='a') {
                    double zav = fin->VS[j].z[i];
                    uint nav = 1;
                    for (jj=j+1; jj<fin->nScan() && 
                             !BP.contained(jj); ++jj) {
                        zav += fin->VS[jj].z[i];
                        ++nav;
                    }
                    S.z.push_back(zav/nav);
                }
            }
            fout.VS.push_back(S);
        }
		
        /** act on x,y **/
        switch (action) {
        case 'A':
            ScaExeAvge(fin, &fout, BP);
            break;
        case 'M':
            ScaExeMerge(fin, &fout, BP);
            break;
        case 'B':
            ScaExeBreak(fin, fout, BP);
            continue;
        default:
            printf("!!! PROG ERROR invalid action\n");
            return;
        }

        NOlm::OloAdd(&fout);
    }
}

void NManip::ScaExeAvge(COld *fin, COld *fout, const CList& BP)
{
    uint iv, ji, jf, jj, i, n, mj;
    vector<double> y;
    string err;
    CScan *Sout;

    for(iv=0; iv<BP.n(); iv++) {
        ji = BP.V[iv];
        jf = ( iv<BP.n()-1 ? BP.V[iv+1] : fin->nScan() );
        // printf("DEBUG iv=%u ji=%u jf=%u\n", iv, ji, jf);
        Sout = &(fout->VS[iv]);
        n = fin->VS[ji].size();
        for (jj=ji; jj<jf; ++jj) {
            if (!fin->VS[jj].x.compatible(
                    fin->VS[ji].x, &err)) {
                printf("! grids %d and %d differ: %s\n",
                       jj, ji, err.c_str());
                return;
            }
        }
        // printf("DEBUG n=%u\n", n);
        Sout->x = fin->VS[ji].x;
        y.assign(n,0.);
        mj = jf-ji;
        // printf("DEBUG mj=%u\n", mj);
        for (jj=ji; jj<jf; jj++)
            for (i=0; i<n; i++)
                y[i] += fin->VS[jj].y[i] / mj;
        Sout->y = y;
    }
}

void NManip::ScaExeMerge(COld *fin, COld *fout, const CList& BP)
{
    uint j, jj, jout, i;
    CScan *Sout;

    for (j=0, jout=0; j<fin->nScan(); ++j) {
        if (!BP.contained(j)) continue;
        Sout = &(fout->VS[jout++]);
        for (jj=j; jj<fin->nScan() && !(jj>j && BP.contained(jj)); ++jj)
            for (i=0; i<fin->VS[jj].size(); ++i)
                Sout->push_back(fin->VS[jj].x[i],
                                fin->VS[jj].y[i]);
    }
}

void NManip::ScaExeBreak(COld *fin, const COld& ftmp, const CList& BP)
{
    COld fout;
    uint j;

    for (j=0; j<fin->nScan(); ) {
        // if (!BP.contained(j)) continue;
        fout = ftmp;
        fout.VS.clear();
        do {
            fout.VS.push_back(fin->VS[j++]);
        } while (j<fin->nScan() && !BP.contained(j));
        NOlm::OloAdd(&fout);
    }
}

void NManip::ScaBreak(void)
{
    COld *fin, ftmp, fout;
    CCoord zco;
    double zval;
    NOlm::IterateD fiter;
    NOlm::SelNew();
    while((fin=fiter())) {
        if (!(fin->ZCo.size())) {
            cout << "! invalid operation / no z coordinate\n";
            return;
        }
        zco = fin->ZCo[0];
        cout << ".. eliminate " << zco << "\n";

        ftmp = *fin;
        ftmp.lDoc.push_back("scans -> files, eliminating " 
                            + zco.str());

        ftmp.ZCo.erase(ftmp.ZCo.begin());

        for (uint j=0; j<ftmp.VS.size(); ) {
            fout = ftmp;
            fout.VS.clear();
            zval = ftmp.VS[j].z[0];
            fout.RPar.push_back( CParam( zco, zval ) );
            ftmp.VS[j].z.erase(ftmp.VS[j].z.begin());
            fout.VS.push_back(ftmp.VS[j]);
            while (++j<ftmp.VS.size() && ftmp.VS[j].z[0]==zval) {
                ftmp.VS[j].z.erase(ftmp.VS[j].z.begin());
                fout.VS.push_back(ftmp.VS[j]);
            }
            NOlm::OloAdd(&fout);
        }
    }
}

void NManip::HistoMake(void)
{
    COld *fin, *f1, fout;
    CScan *sin, *s1, S;
    double sum, center, dx, wgt;
    uint n, nzold, nout;
    static int k1=-1;

    // analyse input data :
    NOlm::IterateD fiter;
    CRange R(2); // start with empty range
    while((fin=fiter()))
        for (uint j=0; j<fin->nScan(); j++)
            R *= fin->VS[j].get_yrange();

    // set grid :
    printf("Input y data range from %g to %g\n", R.inf, R.sup);
    static CGrid G;
    G.Ask("data (input y, output x)");
    printf("DEBUG 1 grid [%s]\n", G.c_str());

    static uint modeN=3;
    modeN = iask("Normalize to unity (+1), to dx (+2) to center (+4)", 
                 modeN);
    if (modeN & 4) {
        cout << "mode desactivated\n";
        return;
    }

    k1 = iask( "Weight from file (-1: unit weight)", k1 );

    // make histograms :
    fiter = NOlm::IterateD();
    NOlm::SelNew();
    while((fin=fiter())) {
        fout = *fin;
        fout.xco = fin->yco;
        fout.yco.name = "h(" + fin->yco.name + ")";
        fout.yco.unit = fin->yco.unit + "-1";
        nzold = fin->ZCo.size();

        if( k1>-1 ){
            if( k1>=NOlm::MOM.size() ){
                printf( "invalid weight file number %i\n", k1 );
                return;
            }
            if( !(f1=NOlm::MOM[k1].D()) ){
                printf("! weight file is not a data file\n");
                return;
            }
            if( f1->nScan()!=fin->nScan() ){
                printf("! weight file differs in number of scans\n");
                return;
            }
        }

        for (uint j=0; j<fin->nScan(); j++) {
            sin = &(fin->VS[j]);
            n = sin->size();
            if( k1>-1 ){
                s1 = &(f1->VS[j]);
                if( s1->size()!=n ){
                    printf( "! weight scan differs in length\n" );
                    return;
                }
            }
            sum = 0;
            center = 0;
            S = *sin;
            nout = G.size();
            S.x = G;
            S.y.clear();
            S.y.resize(nout, 0.);
            for (uint i=0; i<n; i++) {
                wgt = (k1>-1) ? s1->y[i] : 1.0 ;
                S.y[G.pos_force(sin->y[i])] += wgt;
                sum += wgt;
                center += sin->y[i]*wgt;
            }
            if( sum<= 0 ){
                printf( "sum of weights <= 0\n" );
                return;
            }
            if (modeN & 1) 
                for (uint i=0; i<nout; i++)
                    S.y[i] /= sum;
            if (modeN & 2) {
                dx = (S.x[nout-1] - S.x[0]) / nout;
                for (uint i=0; i<nout; i++)
                    S.y[i] /= dx;
            }
            center /= sum;
            if (modeN & 4) 
                for (uint i=0; i<nout; i++) {
                    S.y[i] *= center;
//					S.x[i] /= center;
                }
            fout.VS[j] = S;
        }

        NOlm::OloAdd(&fout);
    }
}

//***************************************************************************//
//* manipulations on files                                                  *//
//***************************************************************************//

void NManip::FilMerge(void)
{
    COld *fin, fout;
    CCoord zco;
    vector<CParam> RPar;    
    vector<bool> RParDiff;    
    NOlm::IterateD fiter;
    NOlm::SelNew();

    if (!(fin=fiter())) {
        printf("! no input files given\n");
        return;
    }

    /** coordinates **/
    CCoord co;
    fiter.Reset();
    co = fiter()->xco;
    while ((fin=fiter())) {
        if (fin->xco != co) {
            printf("! different x coordinates\n");
            return;
        }
    }
    fout.xco = co;
    fiter.Reset();
    co = fiter()->yco;
    while ((fin=fiter())) {
        if (fin->yco != co) {
            printf("! different y coordinates\n");
            return;
        }
    }
    fout.yco = co;
    vector<CCoord> ZCo;
    fiter.Reset();
    ZCo = fiter()->ZCo;
    while ((fin=fiter())) {
        if (fin->ZCo.size() != ZCo.size()) {
            printf("! different numer of z coordinates\n");
            return;
        }
        for (uint i=0; i<ZCo.size(); ++i) {
            if (fin->ZCo[i] != ZCo[i]) {
                printf("! different z%u coordinate\n", i);
                return;
            }
        }
    }
    fout.ZCo = ZCo;

    /** new z coordinate **/
    fout.ZCo.push_back(CCoord("no-in-merge", ""));

    /** check RPar **/
    fiter.Reset();
    RPar = fiter()->RPar;
    RParDiff.resize( RPar.size() );
    for (uint i=0; i<RPar.size(); ++i)
        RParDiff[i] = false;
    while ((fin=fiter())) {
        if (fin->RPar.size() != RPar.size()) {
            printf("! different numer of r parameters\n");
            return;
        }
        for (uint i=0; i<RPar.size(); ++i) {
            if (fin->RPar[i].Co != RPar[i].Co ) {
            printf("! different r parameter\n");
            return;                
            }
        }
        for (uint i=0; i<RPar.size(); ++i) {
            if (fin->RPar[i].val != RPar[i].val ) {
                RParDiff[i] = true;
            }
        }
    }

    for (uint i=0; i<RPar.size(); ++i) {
        if( RParDiff[i] ) {
            fout.ZCo.push_back( RPar[i].Co );
        }
    }

    /** file name **/
    string fnam;
	
    fiter.Reset();
    fnam = fiter()->name;
    while ((fin=fiter())) {
        uint n2=fin->name.size();
        if (n2<fnam.size())
            fnam.erase(fnam.begin()+n2, fnam.end());
        for (uint i=0; i<fnam.size(); ++i)
            if (fnam[i]!=(fin->name)[i])
                fnam[i] = ' ';
    }
    for (string::iterator cp = fnam.begin(); cp != fnam.end(); ++cp)
        if (*cp==' ') fnam.erase(cp--);
	
    if (fnam == "") {
        static uint callNo=0;
        char fnamdef[12];
        sprintf(fnamdef, "merge%03u", callNo++);
        fnam = fnamdef;
    }

    fnam = wask("Joined file name", fnam);
    if (fnam=="") return;
    fout.name = fnam;

    /** doc lines **/
    fout.lDoc.push_back(fnam + " is merger of:");
    fiter.Reset();
    while ((fin=fiter())) {
        fout.lDoc.push_back("> " + fin->name);
        for (uint i=0; i<fin->lDoc.size(); ++i)
            fout.lDoc.push_back("> > " + fin->lDoc[i]);
    }

    /** now merge the data **/
    CScan S;
    fiter.Reset();
    while ((fin=fiter())) {
        for (uint j=0; j<fin->nScan(); j++) {
            S = fin->VS[j];
            S.z.push_back(fiter.index());
            for (uint i=0; i<RPar.size(); ++i) {
                if( RParDiff[i] ) {
                    S.z.push_back( fin->RPar[i].val );
                }
            }
            fout.VS.push_back(S);
        }
    }

    NOlm::OloAdd(&fout);
}

void NManip::FilMergePointwise(void)
{
    COld *fin, fout;
    CScan *Sin, *Sout;
    uint i, j;
    NOlm::IterateD fiter;
    NOlm::SelNew();

    VTEST((!(fin=fiter())), ("! no input files given\n"))

	fout = *fin;

    while ((fin=fiter())) {
        VTEST((fin->nScan() != fout.nScan()), \
              ("!! different # scans\n"));

        VTEST((fin->xco != fout.xco || fin->yco != fout.yco), \
              ("! different x/y coordinates\n"));
		
        for (j=0; j<fout.nScan(); ++j) {
            Sin = &(fin->VS[j]); 
            Sout = &(fout.VS[j]); 
			
            VTEST((Sin->z.size() != Sout->z.size()), \
                  ("!! different # z entries\n"));
            for (i=0; i<Sout->z.size(); ++i) {
                VTEST((Sin->z[i] != Sout->z[i]),
                      ("!! different # z values\n"));
            }

            for (i=0; i<Sin->size(); ++i)
                Sout->push_back(Sin->x[i], Sin->y[i]);
        }

        fout.lDoc.push_back("appended pointwise " + fin->name);
    }

    NOlm::OloAdd(&fout);
}
