//**************************************************************************//
//* FRIDA: fast reliable interactive data analysis                         *//
//* scan.cpp: each online data file contains at least one scan             *//
//* (C) Joachim Wuttke 1990-, v2(C++) 2001-                                *//
//* http://frida.sourceforge.net                                           *//
//**************************************************************************//

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <iostream>
#include <string>
#include <vector>

#include "mystd.h"
#include "range.h"
#include "grid.h"
#include "scan.h"


CScan::CScan(void)
{
}

CScan::CScan(const CGrid& X_in, const vector<double>& Y_in,
	     const vector<double>& Z_in)
{
    if (X_in.size() != Y_in.size()) {
        cout << "CScan::Make/ incompatible input vectors\n";
        Clear();
        return;
    }
    x = X_in;
    y = Y_in;
    z = Z_in;
}

void CScan::push_back(const double x_in, const double y_in)
{
    x.push_back(x_in);
    y.push_back(y_in);
}

void CScan::erase(const uint begin, const uint end)
{
    x.erase(begin, end);
    y.erase(y.begin()+begin, y.begin()+end);
}

uint CScan::size() const
{
    if (x.size()!=y.size()) {
        printf("PROG ERR scan::size #x [%u] <> #y [%u]\n",
               x.size(), y.size());
        return 0;
    }
    return y.size();
}

CRange CScan::get_xrange(const bool logflag) const
{
    double inf=x[0], sup=x[0];
    for (uint i=1; i<size(); i++) {
        if (x[i]<inf) inf=x[i];
        if (x[i]>sup) sup=x[i];
    }
    if (logflag) {
        if (sup<=0) return CRange(2);
        inf = sup;
        for (uint i=0; i<size(); i++) 
            if (x[i]<inf && x[i]>0) inf=x[i];
    }
    return CRange(inf, sup);
}

CRange CScan::get_yrange(const bool logflag) const
{
    return get_yrange(CRange(1), logflag);
}

CRange CScan::get_yrange(const CRange& xsubrange, const bool logflag) const
{
    bool first=1;
    double inf, sup;
    for (uint i=0; i<size(); i++) {
        if (!xsubrange.contained(x[i])) continue;
        if (first) {
            inf=y[i], sup=y[i];
            first = 0;
        } else {
            if (y[i]<inf) inf=y[i];
            if (y[i]>sup) sup=y[i];
        }
    }
    if (first) {
        cout << "no x in subrange\n";
        return CRange(2);
    }
    if (logflag) {
        if (sup<=0) return CRange(2);
        inf = sup;
        for (uint i=0; i<size(); i++) 
            if (xsubrange.contained(x[i]) &&
                y[i]<inf && y[i]>0) inf=y[i];
    }
    return CRange(inf, sup);
}

void CScan::Info(void) const
{
    cout << "This scan contains " << size() << " entries\n";
    for (uint i=0; i<size(); i += size()/16)
        printf("%5d %14f %14f\n", i, x[i], y[i]);
    printf("%5d %14f %14f\n", size()-1, x[size()-1], y[size()-1]);
}

void CScan::Clear(void)
{
    x.clear();
    y.clear();
    z.clear();
}

int CScan::Save( FILE *fd, string fmt ) const
{
    if( fmt=="y08" ){
	for ( uint i=0; i<z.size(); i++ ) 
            fprintf( fd, "    z%u: %18.10g\n", i, z[i] );
	fprintf( fd, "    xy: |2 # %u entries\n", size() );
        for( uint i=0; i<size(); i++ ) 
            fprintf( fd, "      %20.15g %20.15g\n", x[i], y[i] );
    } else if( fmt=="a01" ){
        fprintf( fd, "Z   %u\n", z.size() );
	for ( uint i=0; i<z.size(); i++ ) 
            fprintf( fd, "%18.10g\n", z[i] );
	fprintf( fd, "XY  %u\n", size() );
        for( uint i=0; i<size(); i++ ) 
            fprintf( fd, "%20.15g %20.15g\n", x[i], y[i] );
    }
    return 0;
}

int CScan::Load(FILE *fd, uint nz)
{
    int err;
    uint i, m;
    char lab[4];
    double v, w;

    Clear();

    if ((err=fscanf(fd, "%4c%u\n", lab, &m))!=2) {
        printf("! expecting Z head line, found %d entries\n", err);
        return -1;
    }
    if (strncmp(lab, "Z   ", 4)) {
        printf("! found [%s] in Z head line\n", lab);
        return -1;
    }
    if (m!=nz) {
        printf("! found %u Z entries, expecting %u\n", m, nz);
        return -1;
    }
	
    for (i=0; i<nz; i++) {
        if ((err=fscanf(fd, "%lg\n", &v))!=1) {
            printf("! expecting z value, found %d entries\n", err);
            return -1;
        }
        z.push_back(v);
    }
    // printf(".. read %u z values\n", nz);
		
    if ((err=fscanf(fd, "%4c%u\n", lab, &m))!=2) {
        printf("! expecting XY head line, found %d entries\n", err);
        return -1;
    }
    if (strncmp(lab, "XY  ", 4)) {
        printf("! found [%s] in XY head line\n", lab);
        return -1;
    }
	
    for (i=0; i<m; i++) {
        if ((err=fscanf(fd, "%lg %lg\n", &v, &w))!=2) {
            printf("! expecting xy pair, found %d entries\n", err);
            return -1;
        }
        push_back(v, w);
    }
    // printf(".. read %u xy pairs\n", m);

    return 0;
}

bool CScanSortByZ( const CScan& S1, const CScan& S2 )
{
    if( S1.z.size()!=S2.z.size() )
        return false;
    int nz = S1.z.size();
    for( int iz=0; iz<nz; ++iz ){
        if    ( S1.z[iz]>S2.z[iz] )
            return false;
        if    ( S1.z[iz]<S2.z[iz] )
            return true;
    }
    return false;
}
