#include <string>
#include <vector>
#include <math.h>
using namespace std;
#include "olm.h"
#include "rng.h"
#include "asi.h"
#include "readln.h"
#include "special.h"

const double twopi = 8*atan(1.);
const double m_over_hbar = 1.587e-3; // s/m * A-1
const double k0 = 1.00195, kPG = 0.9364995; // A-1
const double ch_cos = kPG/k0, ch_sin = sqrt( 1-SQR(ch_cos) );

namespace NSpecial {
};


class SimulOut{
public:
    SimulOut( const double x, const double y, const double dE );
    double afx, afy, afz, kfE, // angle and energy in final-beam system
        bfx, bfy, bfz, // angle in fixed-chopper system
        count;
};
    
SimulOut::SimulOut( const double x, const double y, const double dE )
{
    afx = sin(x);
    afy = sin(y);
    afz = sqrt( 1 - SQR(afx) - SQR(afy) );
    bfx = -afx*ch_cos + afz*ch_sin;
    bfz = -afx*ch_sin - afz*ch_cos;
    bfy =  afy;
    kfE = k0+dE;
    count = 0;
}

void NSpecial::SimulateRssm(void) 
{

    // Ask for parameters:
    static int nMC = 100;
    static int nL = 1000;
    static int mosa[3] = {300,300,300};
    static double vel = 0;
    static double sx = 0.09;
    static double sy = 0.03;
    static double dx = 0.03;
    static double dy = 0.03;
    static int nV = 1;
    static int nX = 20;
    static double wX = 0.2;
    static int nY = 2;
    static double wY = 0.1;
    do{ 
        nMC = iask( "# Monte-Carlo steps", nMC );
        nL = iask( "# mosaic layers", nL );
        vel = dask( "chopper speed", vel );
        mosa[0] = iask( "mosaic size in x", mosa[0] );
        mosa[1] = iask( "mosaic size in y", mosa[1] );
        mosa[2] = iask( "mosaic size in z", mosa[2] );
        sx = dask( "mosaic spread in x", sx );
        sy = dask( "mosaic spread in y", sy );
        dx = dask( "incoming divergence in x", dx );
        dy = dask( "incoming divergence in y", dy );
        // nV = iask( "# velocities", nV );
        nX = iask( "# outgoing x angles", nX );
        wX = dask( "outgoing x angle spread", wX );
        nY = iask( "# outgoing y angles", nY );
        wY = dask( "outgoing y angle spread", wY );
    } while( !iask( "ok", 1 ) );
    double kv = m_over_hbar * vel;
    printf( "kv = %f\n", kv );

    // Provisionally hard-coded parameters:
    int debug = 1;
    double cX=0.0, cE=0.0;

    // Prepare recipient array:
    COld olf;
    olf.RPar.push_back( CParam( "totMC", "", nMC*nL ) );
    olf.RPar.push_back( CParam( "mosa-x", "", mosa[0] ) );
    olf.RPar.push_back( CParam( "mosa-y", "", mosa[1] ) );
    olf.RPar.push_back( CParam( "mosa-z", "", mosa[2] ) );
    olf.RPar.push_back( CParam( "diverg-x", "", sx ) );
    olf.RPar.push_back( CParam( "diverg-y", "", sy ) );
    olf.RPar.push_back( CParam( "speed", "m/s", vel ) );
    static CCoord xco, yco;
    olf.name = "simul";
    olf.xco = CCoord("kfx", "");
    olf.yco = CCoord("cnts", "");
//    olf.ZCo.push_back(CCoord("vel", ""));
    olf.ZCo.push_back(CCoord("kfx", ""));
    vector<SimulOut> kf;
    for( int jY=0; jY<nY; ++jY ){
        for( int jX=0; jX<nX; ++jX ){
//            for( int jV=0; jV<nV; ++jV ){
                kf.push_back( SimulOut( nX>1 ? -wX+cX+2*wX*jX/(nX-1) : 0.0,
                                        nY>1 ? wY*jY/(nY-1) : 0.0,
                                        0.0 ) );
//            }
        }
    }
    int nkf = kf.size();

    // Monte-Carlo loop:
    for( int iMC=0; iMC<nMC; ++iMC ){
        double scattered = 0;
        double wgt = 1.0/nMC/nL; // /mosa[0]/mosa[1]/mosa[2];
        for( int itau=0; itau<nL; ++itau ){
            // initialize neutron
            double aix = NRNG::uniform( -dx, dx );
            double aiy = NRNG::uniform( -dy, dy );
            double aiz = sqrt( 1 - SQR(aix) - SQR(aiy) );
            // transform into chopper frame
            double bix =  aix*ch_cos + aiz*ch_sin;
            double biz = -aix*ch_sin + aiz*ch_cos;
            double biy =  aiy;
            // track neutron through mosaic layers
            double tx = 2*kPG*sx*NRNG::gaussian( 1.0, 4.0 );
            double ty = 2*kPG*sy*NRNG::gaussian( 1.0, 4.0 );
            double tz = sqrt( SQR(2*kPG) - tx*tx - ty*ty );
            for( int ikf=0; ikf<nkf; ++ikf ){
                SimulOut *out = &(kf[ikf]);
                double kiE = -bix*kv +
                    sqrt( SQR(bix*kv) + SQR(out->kfE) + 2*out->bfx*kv );
                double trans = 1;
                double qa[3];
                qa[0] = (kiE*bix - out->kfE*out->bfx) - tx;
                qa[1] = (kiE*biy - out->kfE*out->bfy) - ty;
                qa[2] = (kiE*biz - out->kfE*out->bfz) - tz;
                printf( "%3i %12.7f %12.7f %12.7f  %12.7f %12.7f %12.7f\n", 
                        ikf,
                        kiE*bix - out->kfE*out->bfx,
                        kiE*biy - out->kfE*out->bfy,
                        kiE*biz - out->kfE*out->bfz,
                        tx, ty, tz );
                for( int ia=0; ia<3; ++ia ){
                    double arg = mosa[ia]*qa[ia];
                    trans *= mosa[ia] *
                        ( fabs(arg)<1.e-20 ? 1.0 : SQR( sin( arg ) / arg ) );
                }
                out->count += wgt*trans;
                scattered += wgt*trans;
            }
        }
        printf( "%4i scattered %9.6g\n", iMC, scattered );
    }

    // Save position-dependent intensities:
    for( int j=0; j<nY; ++j ){
        CScan S;
        S.z.push_back( kf[j*nX].afy );
        for( int i=0; i<nX; ++i ){
            int ikf = j*nX+i;
            SimulOut *out = &(kf[ikf]);
            S.push_back( out->afx, out->count );
        }
        olf.VS.push_back( S );
    }
    NOlm::SelNew();
    NOlm::OloAdd(&olf);
}

void altlast()
{
    // Read tau values:
    /*
    string fnam = "/home/jwu/rs/mosaic/tau0.xy";
    FILE *ft;
    if( !(ft = fopen( fnam.c_str(), "r" ) ) ){
        printf( "cannot open %s\n", ft );
    }
    vector<double> taux, tauy, tauz;
    double tx, ty, tz;
    while( !( fscanf( ft, "%lf %lf\n", &tx, &ty )==EOF ) ){
        tx *= sx;
        ty *= sy;
        tz = sqrt( 1 - tx*tx - ty*ty );
        taux.push_back( amp*tx );
        tauy.push_back( amp*ty );
        tauz.push_back( amp*tz );
    }
    fclose( ft );
    int ntau = taux.size();
    printf( "read %i mosaic orientations\n", ntau );
    */
}
